# Copyright 2013 Cisco Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module generated the python script using UCS GUI, GUI logs and xml
request.
"""

from __future__ import print_function

import os
import sys
import time
import glob
import re
import xml.dom
import xml.dom.minidom
from os.path import dirname
from .. import ucsgenutils
from .. import ucscoreutils
from ..ucsconstants import Status, NamingPropertyId, YesOrNo
from ..ucsexception import UcsValidationException

import logging

log = logging.getLogger('ucs')

# variable declaration
_display_xml = False
_outfile_flag = False
_outfile_path = None
_outfile = None
_and_count = 0
_or_count = 0
_not_count = 0
_multi_line_method = ["<configConfMo", "<lsInstantiateNNamedTemplate",
                      "<statsClearInterval", "<lsClone", "<lsTemplatise",
                      "<lsInstantiateTemplate", "<lsInstantiateNTemplate",
                      "<configConfRename"]
_single_line_method = ["</configConfMo", "</lsInstantiateNNamedTemplate",
                       "</statsClearInterval", "</lsClone", "</lsTemplatise",
                       "</lsInstantiateTemplate", "</lsInstantiateNTemplate",
                       "/>", "</configConfRename"]


# class declaration
class _ClassStatus:
    """enum for mo status."""
    NONE_ = 0
    CREATED = 1
    MODIFIED = 2
    REMOVED = 4
    DELETED = 8
    GET = 16


# Function Definition
# ---- START - OF - GENERIC - FUCNTION ----
def _get_class_id_for_dn(dn):
    """
    Internal method to get the class id for a given dn
    """

    rns = dn.split('/')
    class_id = None
    parent_class_id = None

    for rn in rns:
        class_id = _get_class_id_for_rn(rn, parent_class_id)
        if class_id is None:
            break
        parent_class_id = class_id

    return class_id


def _get_class_id_for_rn(rn, prev_class_id=None):
    """
    Internal method to get the class id for a given rn
    """

    if not prev_class_id:
        prev_class_id = "TopRoot"

    meta_class_id = ucscoreutils.find_class_id_in_mo_meta_ignore_case(
        prev_class_id)
    if meta_class_id is None:
        raise UcsValidationException(
            '[Error]: class_id [%s] is not valid' % (
                prev_class_id))

    mo_meta = ucscoreutils.get_mo_property_meta(meta_class_id, "mo_meta")
    if mo_meta is None:
        raise UcsValidationException(
            '[Error]: mo_meta for class_id [%s] is not valid'
            % prev_class_id)

    for child_class_id in mo_meta.children:
        # 1.If Rn does not contain [, then there is no naming property.
        #   Check directly.
        # 2. If Rn contains [,
        # 2.1) Check if the Rn contains a "-" and ChildClassId.Rn does not,
        #      then proceed to next child_class_id.
        #      substitute \[[^\]]*\] with .* and do a regex match
        # 2.2) If Rn does not contain a "-",
        #      then probably the child_class_id has only the naming property
        #      without any suffix or prefix. Check for "[*]".
        #      Need to see, if there is a possibility of more than one [*]
        #      directly under a ClassId

        # child_class_id = "orgorg"
        child_class_id = ucscoreutils.find_class_id_in_mo_meta_ignore_case(
            child_class_id)
        child_mo_meta = ucscoreutils.get_mo_property_meta(child_class_id,
                                                          "mo_meta")
        if child_mo_meta is None:
            continue

        # 2. Check if there is a naming property
        match = re.search(r"(\[[^\]]+\])", child_mo_meta.rn)
        if not match:
            if child_mo_meta.rn == rn:
                return child_mo_meta.name
            continue

        # 2.1 Check if it has a prefix or suffix
        match = re.search(r"(([^\]]\[)|(\][^\[]))", child_mo_meta.rn)
        if match:
            mod_meta_rn = re.sub(r"\[([^\]]+)\]", r"(?P<\1>.*?)",
                                 child_mo_meta.rn)
            if re.match(r"\|", mod_meta_rn):
                mod_meta_rn = re.sub(r"\|", r"\|", mod_meta_rn)
            pattern = "^" + mod_meta_rn + "$"
            if re.match(pattern, rn):
                return child_mo_meta.name
        else:
            continue

    for child_class_id in mo_meta.children:
        child_mo_meta = ucscoreutils.get_mo_property_meta(child_class_id,
                                                          "mo_meta")
        if child_mo_meta is None:
            continue
        match = re.match(r"^(\[[^\]]+\])+$", child_mo_meta.rn)
        if match:
            return child_mo_meta.name

    return None


def _get_prop_name(prop):
    """
    Internal method to modify property name as python convention
    """

    new_prop = re.sub('_+', '_',
                      re.sub('^_', '',
                             re.sub(r'[/\-: +]', '_',
                                    re.sub(r'([a-z0-9])([A-Z])',
                                           r'\g<1>_\g<2>', prop)))).upper()
    return new_prop


def _first_capital(string):
    """
    Internal method to make first letter capital
    """

    string = string[:1].upper() + string[1:]
    return string


def _is_root_node(dom, tag_name):
    """
    Internal method to check if it root xml node
    """

    root_node_tag_name = dom.documentElement.tag_name
    if root_node_tag_name == tag_name:
        return True
    else:
        return False


def _get_pair_nodes(root_node):
    """
    Internal method to get "pair" nodes under root_node
    """

    method_elem = root_node
    in_configs_elem_list = method_elem.getElementsByTagName("inConfigs")
    in_configs_elem = in_configs_elem_list[0]
    pair_elems_list = in_configs_elem.getElementsByTagName("pair")
    return pair_elems_list


def _get_only_elem_child_node(node):
    """
    Internal method to return only first element child node
    """

    child_list = [child_node for child_node in node.childNodes
                  if child_node.nodeType == child_node.ELEMENT_NODE]
    return child_list[0]


def _get_elem_child_nodes(node):
    """
    Internal method to return all element child nodes
    """

    child_list = [child_node for child_node in node.childNodes if
                  child_node.nodeType == child_node.ELEMENT_NODE]
    return child_list


def _dump_xml_on_screen(doc):
    """
    Internal method to display xml extracted to generate sdk code.
    """

    global _outfile_path, _outfile_flag
    xml_new = doc.toprettyxml(indent=" ")
    xml_new = re.sub(r"^.*?xml version.*\n", "", xml_new)
    xml_new = re.sub(r"\n[\s]*\n", "\n", xml_new)
    multiline_pattern = re.compile(r"^(.*)", flags=re.MULTILINE)
    xml_new = re.sub(multiline_pattern, r"#\1", xml_new)
    if _outfile_flag:
        _outfile = open(_outfile_path, 'a')
        _outfile.write("\n##### Start-Of-XML #####\n")
        _outfile.write("%s" % xml_new)
        _outfile.write("\n##### End-Of-XML #####\n")
        _outfile.close()
    else:
        print("\n##### Start-Of-XML #####\n")
        print(xml_new)
        print("\n##### End-Of-XML #####\n")


def _check_if_any_list_value_in_string(listx, line):
    """
    Internal method to test if any of the list value present in a line.
    """

    flag = False
    for value in listx:
        if value in line:
            flag = True
            break
    return flag


def _create_property_map_from_node(class_node, class_status):
    gmo_flag = False
    property_map = {}

    if ucscoreutils.find_class_id_in_mo_meta_ignore_case(
            class_node.localName) is None:
        gmo_flag = True

    if not gmo_flag:
        peer_class_id = _first_capital(class_node.localName)
        peer_class_ref = ucscoreutils.load_class(peer_class_id)
        peer_class_naming_props = peer_class_ref.naming_props
        peer_class_prop_map = peer_class_ref.prop_map
        peer_class_prop_meta = peer_class_ref.prop_meta
        peer_class_mo_meta = peer_class_ref.mo_meta

    # create property map for attributes
    attributes_dict = dict(class_node.attributes.items())
    if "dn" in attributes_dict:
        prop_dn = attributes_dict["dn"]
        prop_rn = os.path.basename(prop_dn)
    elif "rn" in attributes_dict:
        prop_rn = attributes_dict["rn"]

    naming_props_py_from_rn = ucscoreutils.get_naming_props(
        prop_rn, peer_class_mo_meta.rn)

    for attr, val in class_node.attributes.items():
        name = attr
        value = val

        if name in ['dn', 'rn', 'status']:
            continue

        # prop_ucs : ucs naming convention
        # prop_py  : python naming convention
        # if name.lower() == "Filter".lower():
        #     prop_ucs = "FilterValue"
        # else:
        #     prop_ucs = name
        prop_ucs = name

        if not gmo_flag and prop_ucs in peer_class_prop_map:
            prop_py = peer_class_prop_map[prop_ucs]
            prop_meta = peer_class_prop_meta[prop_py]
            prop_access = prop_meta.access
            if prop_access in [2, 4]:
                continue
            elif prop_access in [0, 1]:
                if class_status & _ClassStatus.CREATED != \
                        _ClassStatus.CREATED:
                    continue
            if not prop_meta.validate_property_value(val):
                continue
        else:
            prop_py = prop_ucs

        if "\\" in val:
            value = 'r"' + str(val) + '"'
        else:
            value = '"' + str(val) + '"'

        property_map[prop_py] = value

    if class_status & _ClassStatus.CREATED == _ClassStatus.CREATED:
        for naming_prop_ucs in peer_class_naming_props:
            naming_prop_py = peer_class_prop_map[naming_prop_ucs]
            if naming_prop_py not in property_map:
                val = naming_props_py_from_rn[naming_prop_py]
                if peer_class_prop_meta[naming_prop_py]. \
                        validate_property_value(val):
                    if "\\" in val:
                        value = 'r"' + str(val) + '"'
                    else:
                        value = '"' + str(val) + '"'

                    property_map[naming_prop_py] = value

    return property_map


# ---- END - OF - GENERIC - FUCNTION ----

# ---- START - OF - METHOD SPECIFIC - FUNCTION ----
def _get_config_conf_cmdlet(node, is_pair_node):
    """
    Internal method to process configConf request.
    """

    cmdlet = ""
    if node is None:
        return None
    class_node = None
    key = ""

    if is_pair_node:
        key = node.getAttribute(NamingPropertyId.KEY)
        class_node = _get_only_elem_child_node(node)
        if class_node is None:
            return None
    else:
        key = node.getAttribute(NamingPropertyId.DN)
        class_node = node

    dn = ""
    mo_tag = ""
    if class_node.hasAttribute(NamingPropertyId.DN):
        dn = class_node.getAttribute(NamingPropertyId.DN)

    if class_node.hasAttribute(
        NamingPropertyId.STATUS) and \
            class_node.getAttribute(NamingPropertyId.STATUS) is not \
            None:
        mo_tag = "mo"
    else:
        if class_node.hasChildNodes() and len(
                _get_elem_child_nodes(class_node)) > 0:
            mo_tag = "obj"
        else:
            mo_tag = "mo"

    import_list = []
    obj_flag = False

    top_cmdlet, op_flag, import_list = _form_configconf_cmdlet(
        class_node=class_node,
        key=key,
        tag=mo_tag,
        import_list=import_list)

    if class_node.hasChildNodes() and \
            len(_get_elem_child_nodes(class_node)) > 0:
        call_count = 1
        cmdlet = ""
        for child_node in _get_elem_child_nodes(class_node):
            sub_cmdlet, import_list, obj_flag = _get_config_conf_sub_cmdlet(
                child_node,
                dn, mo_tag,
                call_count,
                import_list,
                obj_flag)

            call_count += 1
            if sub_cmdlet is not None:
                cmdlet += "\n" + sub_cmdlet
            else:
                call_count -= 1

    if op_flag in ["add", "addmodify", "set", "remove"] and mo_tag == "mo":
        if op_flag == "add":
            cmdlet += "\n" + "handle.add_mo(mo)\n"
        elif op_flag == "addmodify":
            cmdlet += "\n" + "handle.add_mo(mo, True)\n"
        elif op_flag == "set":
            cmdlet += "\n" + "handle.set_mo(mo)\n"
        elif op_flag == "remove":
            cmdlet += "handle.remove_mo(mo)"

    import_cmdlet = ""
    for mo in import_list:
        mo_folder = re.match("([a-z])+", ucsgenutils.word_l(mo)).group()
        import_cmdlet += "from ucsmsdk.mometa.%s.%s import %s\n" % (
            mo_folder, mo, mo)

    main_cmdlet = ""
    if import_cmdlet:
        main_cmdlet += "\n" + import_cmdlet
    if mo_tag == "obj" and not obj_flag:
        top_cmdlet = ""

    main_cmdlet += "\n" + top_cmdlet
    main_cmdlet += cmdlet

    return main_cmdlet


def _get_config_conf_sub_cmdlet(class_node, parent_dn, parent_mo_tag,
                                parent_call_count, import_list, obj_flag):
    """
    Internal method to process configConf request.
    """

    cmdlet = ""
    dn = ""
    # use_parent_mo = False ##if the parent mo should be used at this level

    if class_node.hasAttribute(NamingPropertyId.DN):
        dn = class_node.getAttribute(NamingPropertyId.DN)
    elif class_node.hasAttribute(NamingPropertyId.RN):
        dn = parent_dn + "/" + class_node.getAttribute(NamingPropertyId.RN)

    count = 1
    if parent_mo_tag == "obj":
        tag = "mo"
    else:
        tag = parent_mo_tag + "_" + str(parent_call_count)

    cmdlet, op_flag, import_list = _form_configconf_cmdlet(
        class_node=class_node,
        key=dn,
        tag=tag,
        import_list=import_list,
        parent_tag=parent_mo_tag,
        sub_cmdlet=True)

    # Recursively cater to subnodes
    for child_node in _get_elem_child_nodes(class_node):
        sub_cmdlet, import_list, obj_flag = _get_config_conf_sub_cmdlet(
            child_node,
            dn,
            tag,
            count,
            import_list,
            obj_flag)
        count += 1
        if sub_cmdlet is not None:
            cmdlet += "\n" + sub_cmdlet
        else:
            count -= 1

    # if op_flag in ["add", "addmodify", "set", "remove"] and tag == "mo":
    if op_flag == "add" and tag == "mo":
        cmdlet += "\n" + "handle.add_mo(mo)\n"
    if op_flag == "addmodify" and tag == "mo":
        cmdlet += "\n" + "handle.add_mo(mo, True)\n"
    if op_flag == "set":
        cmdlet += "\n" + "handle.set_mo(mo)\n"
    if op_flag == "remove":
        cmdlet += "handle.remove_mo(%s)\n" % tag

    if not obj_flag:
        if parent_mo_tag == "obj" and op_flag in ["add", "addmodify", "set"]:
            obj_flag = True

    return cmdlet, import_list, obj_flag


def _form_configconf_cmdlet(class_node, key, tag, import_list, parent_tag=None,
                            sub_cmdlet=False):
    """
    Internal method to process configConf request.
    """

    cmdlet = ""
    class_status = _ClassStatus.NONE_
    op_flag = None
    parent_dn = dirname(key)
    peer_class_id = _first_capital(class_node.localName)

    parent_class_id = _get_class_id_for_dn(parent_dn)
    if parent_class_id is None:
        parent_class_id = ""

    if class_node.hasAttribute(
        NamingPropertyId.STATUS) and \
            class_node.getAttribute(NamingPropertyId.STATUS) is not \
            None:
        cs_list = []
        cs_list = class_node.getAttribute(NamingPropertyId.STATUS).split(',')
        cs_list = [x.strip() for x in cs_list]
        if Status.CREATED in cs_list:
            class_status |= _ClassStatus.CREATED
        if Status.MODIFIED in cs_list:
            class_status |= _ClassStatus.MODIFIED
        if Status.DELETED in cs_list:
            class_status |= _ClassStatus.DELETED
        if Status.REMOVED in cs_list:
            class_status |= _ClassStatus.REMOVED
    else:
        if sub_cmdlet:
            class_status = _ClassStatus.CREATED | _ClassStatus.MODIFIED
        else:
            if class_node.hasChildNodes() and \
                    len(_get_elem_child_nodes(class_node)) > 0:
                class_status = _ClassStatus.GET
            else:
                class_status = _ClassStatus.CREATED | _ClassStatus.MODIFIED

    property_map = _create_property_map_from_node(class_node, class_status)

    parent_mo_or_dn = '"' + str(parent_dn) + '"'
    if sub_cmdlet:
        parent_mo_or_dn = parent_tag

    # make cmdlet
    if class_status & _ClassStatus.GET == _ClassStatus.GET:
        op_flag = "get"
        cmdlet = "%s = handle.query_dn(\"%s\")" % (tag, key)
    elif class_status & _ClassStatus.DELETED == _ClassStatus.DELETED or \
            class_status & \
            _ClassStatus.REMOVED == _ClassStatus.REMOVED:
        op_flag = "remove"
        cmdlet = "%s = handle.query_dn(\"%s\")\n" % (tag, key)
    elif class_status & _ClassStatus.CREATED == _ClassStatus.CREATED:
        op_flag = "add"
        if peer_class_id not in import_list:
            import_list.append(peer_class_id)
        if class_status & _ClassStatus.MODIFIED == _ClassStatus.MODIFIED:
            op_flag = "addmodify"
            add_prop_map = ", ".join(
                [k + "=" + v for k, v in ucsgenutils.iteritems(property_map)])
            cmdlet = '%s = %s(parent_mo_or_dn=%s, %s)' % (
                tag, peer_class_id, parent_mo_or_dn, add_prop_map)
        else:
            add_prop_map = ", ".join(
                [k + "=" + v for k, v in ucsgenutils.iteritems(property_map)])
            cmdlet = '%s = %s(parent_mo_or_dn=%s, %s)' % (
                tag, peer_class_id, parent_mo_or_dn, add_prop_map)
    elif class_status & _ClassStatus.MODIFIED == _ClassStatus.MODIFIED:
        op_flag = "set"
        set_prop_map = "\n".join(
            [tag + "." + k + " = " + v for k, v in ucsgenutils.iteritems(property_map)])
        cmdlet = "%s = handle.query_dn(\"%s\")\n%s\n" % (
            tag, key, set_prop_map)
    else:
        print("Throw Exception XML request status (%s) is invalid." % (
            class_node.getAttribute(NamingPropertyId.STATUS)))

    return cmdlet, op_flag, import_list


def _generate_config_conf_cmdlets(xml_string):
    """
    Internal method which takes xmlstring, and generate script for configConfMo
    and configConfMos methods.
    """

    doc = xml.dom.minidom.parseString(xml_string)
    top_node = doc.documentElement

    if top_node is None:
        return

    cmdlet = ""

    if len(top_node.getElementsByTagName("inConfigs")) != 0:
        pair_nodes = _get_elem_child_nodes(
            top_node.getElementsByTagName("inConfigs")[0])
    else:
        pair_nodes = None

    if pair_nodes is None or len(pair_nodes) < 1:
        node = top_node.getElementsByTagName("inConfig")[0]
        if node is None:
            return

        node = _get_only_elem_child_node(node)
        if node is None:
            return
        cmdlet = _get_config_conf_cmdlet(node, False)
    elif len(pair_nodes) > 1:
        temp_cmdlet = ""
        temp_dn = ""
        temp_mo = ""
        count = 0
        dict_mos = {}

        # cmdlet = "handle.start_ucs_transaction()" + "\n"
        for node in pair_nodes:
            temp_dn = node.getAttribute(NamingPropertyId.KEY)
            temp_mo = "mo"
            if count > 0:
                temp_mo += str(count)
            if temp_dn not in dict_mos:
                dict_mos[temp_dn] = temp_mo

            # check if parent is already there in dictionary
            child_dn = os.path.basename(temp_dn)
            parent_dn = os.path.dirname(temp_dn)
            if parent_dn in dict_mos:
                node.setAttribute(NamingPropertyId.KEY, child_dn)
                temp_cmdlet = _get_config_conf_cmdlet(node, True)
                cmdlet += temp_mo + " = " + dict_mos[
                    parent_dn] + " | " + temp_cmdlet + "\n"
            else:
                temp_cmdlet = _get_config_conf_cmdlet(node, True)
                cmdlet += temp_cmdlet + "\n"

            count += 1

            # cmdlet += "handle.complete_ucs_transaction()"
    else:
        cmdlet = _get_config_conf_cmdlet(pair_nodes[0], True)

    cmdlet += "\nhandle.commit()"

    if cmdlet is not None:
        return cmdlet


def _generate_config_resolve_cmdlet(xml_string, method):
    """
    Internal method which takes xmlstring, and generate script for
    configResolveDn, configResolveDns, configResolveClass and
    configResolveClasses methods.
    """

    doc = xml.dom.minidom.parseString(xml_string)
    cmdlet = ""
    if not _is_root_node(doc, method):
        return

    # <configResolveDn>
    if method == "configResolveDn":
        top_node = doc.documentElement
        if top_node is None:
            return
        dn = top_node.getAttribute(NamingPropertyId.DN)
        in_hierarchical = ""
        if top_node.hasAttribute("in_hierarchical"):
            in_hierarchical = top_node.getAttribute("in_hierarchical")
        if in_hierarchical.lower() == "true":
            in_hierarchical_value = YesOrNo.TRUE
        else:
            in_hierarchical_value = YesOrNo.FALSE

        cmdlet = 'handle.config_resolve_dn("%s", "%s")' % (
            dn, in_hierarchical_value)
    # <configResolveDns>
    elif method == "configResolveDns":
        top_node = doc.documentElement
        if top_node is None:
            return
        in_hierarchical = ""
        if top_node.hasAttribute("in_hierarchical"):
            in_hierarchical = top_node.getAttribute("in_hierarchical")

        dn_nodes = _get_elem_child_nodes(
            top_node.getElementsByTagName("inDns")[0])
        if dn_nodes is None:
            return

        cmdlet = "dn_set = DnSet()" + "\n"
        temp_dn = ""

        for node in dn_nodes:
            temp_dn = node.getAttribute(NamingPropertyId.VALUE)
            cmdlet += "dn = Dn()\ndn.attr_set(\"value\"," \
                      "\"%s\")\ndn_set.child_add(dn)\n" % temp_dn

        if in_hierarchical.lower() == "true":
            in_hierarchical_value = YesOrNo.TRUE
        else:
            in_hierarchical_value = YesOrNo.FALSE

        cmdlet += 'handle.config_resolve_dns(dn_set, "%s")' % \
                  in_hierarchical_value
    # <configResolveClass>
    elif method == "configResolveClass":
        _and_count = 0
        _or_count = 0
        _not_count = 0
        top_node = doc.documentElement
        filter_node = _get_only_elem_child_node(top_node)
        print(filter_node)
        if top_node is None or filter_node is None:
            return

        in_hierarchical = ""
        if top_node.hasAttribute("in_hierarchical") is not None:
            in_hierarchical = top_node.getAttribute("in_hierarchical")

        class_id = ""
        if top_node.hasAttribute("class_id") is not None:
            class_id = top_node.getAttribute("class_id")

        if in_hierarchical.lower() == "true":
            in_hierarchical_value = YesOrNo.TRUE
        else:
            in_hierarchical_value = YesOrNo.FALSE

        cmdlet = "in_filter = FilterFilter()\n" + _create_python_filter_code(
            filter_node,
            "inFilter") + 'handle.config_resolve_class(' \
                          '"%s", in_filter, "%s")' % (class_id,
                                                      in_hierarchical_value)
    # <configResolveClasses>
    elif method == "configResolveClasses":
        top_node = doc.documentElement
        if top_node is None:
            return

        in_hierarchical = ""
        if top_node.hasAttribute("in_hierarchical") is not None:
            in_hierarchical = top_node.getAttribute("in_hierarchical")

        class_id_nodes = _get_elem_child_nodes(
            top_node.getElementsByTagName("inIds")[0])

        cmdlet = '\nfrom ucsmsdk.ucsbasetype import ClassIdSet, ClassId\n'
        cmdlet += "id_set = ClassIdSet()" + "\n"
        temp_class_id = ""
        for node in class_id_nodes:
            temp_class_id = node.getAttribute(NamingPropertyId.VALUE)
            cmdlet += "class_id = ClassId()\nclass_id.attr_set(\"value\"," \
                      "\"%s\")\nidSet.child_add(class_id)\n" % temp_class_id

        if in_hierarchical.lower() == "true":
            in_hierarchical_value = YesOrNo.TRUE
        else:
            in_hierarchical_value = YesOrNo.FALSE

        cmdlet += 'handle.config_resolve_classes(id_set, "%s")' % \
                  in_hierarchical_value

    return cmdlet


def _create_python_filter_code(parent_node, parent_filter_name):
    """
    Internal method to provide filter support.
    """

    cmdlet = ""
    filter_name = ""
    temp_name = ""

    for node in _get_elem_child_nodes(parent_node):
        if node.localName == "and":
            temp_name = "andFilter" + str(_and_count)
            _and_count += 1
            cmdlet = temp_name
            cmdlet += " = AndFilter()\n"
            cmdlet += _create_python_filter_code(node, temp_name)
            cmdlet += parent_filter_name
            cmdlet += ".child_add(" + temp_name + ")\n"
            continue

        if node.localName == "or":
            temp_name = "orFilter" + str(_or_count)
            _or_count += 1
            cmdlet = temp_name
            cmdlet += " = OrFilter()\n"
            cmdlet += _create_python_filter_code(node, temp_name)
            cmdlet += parent_filter_name
            cmdlet += ".child_add(" + temp_name + ")\n"
            continue

        if node.localName == "not":
            temp_name = "notFilter" + str(_not_count)
            _not_count += 1
            cmdlet = temp_name
            cmdlet += " = NotFilter()\n"
            cmdlet += _create_python_filter_code(node, temp_name)
            cmdlet += parent_filter_name
            cmdlet += ".child_add(" + temp_name + ")\n"
            continue

        if node.localName == "eq":
            filter_name = "eqFilter"
        if node.localName == "ne":
            filter_name = "neFilter"
        if node.localName == "gt":
            filter_name = "gtFilter"
        if node.localName == "lt":
            filter_name = "ltFilter"
        if node.localName == "le":
            filter_name = "leFilter"
        if node.localName == "wcard":
            filter_name = "wcardFilter"
        if node.localName == "anybit":
            filter_name = "anybitFilter"
        if node.localName == "allbits":
            filter_name = "allbitsFilter"
        if node.localName == "bw":
            filter_name = "bwFilter"

        cmdlet += filter_name + " = " + _first_capital(filter_name) + "()\n"

        for name, value in node.attributes.items():
            cmdlet += "%s.%s = \"%s\"\n" % (
                filter_name, _first_capital(name), value)

        cmdlet += parent_filter_name + ".child_add(" + filter_name + ")\n"

    return cmdlet


def _generate_single_clone_cmdlets(xml_string, is_template):
    """
    Internal method which takes xmlstring, and generate script for
    lsClone and lsInstantiateTemplate methods.
    """

    doc = xml.dom.minidom.parseString(xml_string)
    node = None
    top_node = doc.documentElement
    if top_node is None:
        return

    if is_template:
        if top_node.localName == "lsInstantiateTemplate":
            node = top_node
    else:
        if top_node.localName == "lsClone":
            node = top_node

    dn = ""
    if node.hasAttribute(NamingPropertyId.DN):
        dn = node.getAttribute(NamingPropertyId.DN)
    else:
        print("Attribute 'Dn' not available")  # writewarning in dotnet
        return

    sp_new_name = ""
    if node.hasAttribute("inServerName"):
        sp_new_name = node.getAttribute("inServerName")
    else:
        print("Attribute 'inServerName' not available")
        return

    dest_org = ""
    if node.hasAttribute("inTargetOrg"):
        dest_org = node.getAttribute("inTargetOrg")

    match_object = re.match(
        r"^(org-[\-\.:_a-zA-Z0-9]{1,16}/)*org-[\-\.:_a-zA-Z0-9]{1,16}", dn)
    if match_object is not None:
        # source_org = match_object.group(0)
        pass
    else:
        print("Attribute 'Dn' is corrupt")
        return

    cmdlet = ""
    in_hierarchical = ""
    if node.hasAttribute("in_hierarchical") is not None:
        in_hierarchical = node.getAttribute("in_hierarchical")

    if in_hierarchical.lower() == "true":
        in_hierarchical_value = YesOrNo.TRUE
    else:
        in_hierarchical_value = YesOrNo.FALSE

    in_error_on_existing = ""
    if node.hasAttribute("inErrorOnExisting") is not None:
        in_error_on_existing = node.getAttribute("inErrorOnExisting")

    if is_template:
        cmdlet = '\nfrom ucsmsdk.ucsmethodfactory ' \
                 'import ls_instantiate_template\n'
        cmdlet += 'elem = ls_instantiate_template(cookie=' \
                  'handle.cookie, dn="%s", in_error_on_existing=' \
                  '"%s", in_server_name="%s", in_target_org=' \
                  '"%s", in_hierarchical="%s")' % (dn,
                                                   in_error_on_existing,
                                                   sp_new_name,
                                                   dest_org,
                                                   in_hierarchical_value)
        cmdlet += '\nmo_list = handle.process_xml_elem(elem)\n'
    else:
        cmdlet = '\nfrom ucsmsdk.ucsmethodfactory import ls_clone\n'
        cmdlet += 'elem = ls_clone(cookie=handle.cookie, dn=' \
                  '"%s", in_server_name="%s", in_target_org=' \
                  '"%s", in_hierarchical="%s")' % (dn,
                                                   sp_new_name,
                                                   dest_org,
                                                   in_hierarchical_value)
        cmdlet += '\nmo_list = handle.process_xml_elem(elem)\n'

    return cmdlet


def _generate_ls_templatise_cmdlets(xml_string):
    """
    Internal method which takes xmlstring, and generate script for
    lsTemplatise method.
    """

    doc = xml.dom.minidom.parseString(xml_string)
    node = None
    top_node = doc.documentElement
    if top_node is None:
        return

    if top_node.localName == "lsTemplatise":
        node = top_node
    else:
        print("Check if Method is <lsTemplatise>")
        return

    dn = ""
    if node.hasAttribute(NamingPropertyId.DN):
        dn = node.getAttribute(NamingPropertyId.DN)
    else:
        print("Attribute 'Dn' not available")  # writewarning in dotnet
        return

    sp_new_name = ""
    if node.hasAttribute("inTemplateName"):
        sp_new_name = node.getAttribute("inTemplateName")
    else:
        print("Attribute 'inTemplateName' not available")
        return

    template_type = ""
    if node.hasAttribute("inTemplateType"):
        template_type = node.getAttribute("inTemplateType")
    else:
        print("Attribute 'inTemplateType' not available")
        return

    dest_org = ""
    if node.hasAttribute("inTargetOrg"):
        dest_org = node.getAttribute("inTargetOrg")

    match_object = re.match(
        r"^(org-[\-\.:_a-zA-Z0-9]{1,16}/)*org-[\-\.:_a-zA-Z0-9]{1,16}", dn)
    if match_object is not None:
        # source_org = match_object.group(0)
        pass
    else:
        print("Attribute 'Dn' is corrupt")
        return

    cmdlet = ""
    in_hierarchical = ""
    if node.hasAttribute("in_hierarchical") is not None:
        in_hierarchical = node.getAttribute("in_hierarchical")
    if in_hierarchical.lower() == "true":
        in_hierarchical_value = YesOrNo.TRUE
    else:
        in_hierarchical_value = YesOrNo.FALSE

    if dest_org is not None:
        cmdlet = '\nfrom ucsmsdk.ucsmethodfactory import ls_templatise\n'
        cmdlet += 'elem = ls_templatise(cookie=handle.cookie,' \
                  'dn="%s", in_target_org="%s", in_template_name="%s", ' \
                  'in_template_type="%s", in_hierarchical="%s")' % (dn,
                                                                    dest_org,
                                                                    sp_new_name,
                                                                    template_type,
                                                                    in_hierarchical_value)
        cmdlet += '\nmo_list = handle.process_xml_elem(elem)\n'
    else:
        cmdlet = '\nfrom ucsmsdk.ucsmethodfactory import ls_templatise\n'
        cmdlet += 'elem = ls_templatise(cookie=handle.cookie,' \
                  'dn="%s", in_target_org="org-root", ' \
                  'in_template_name="%s", ' \
                  'in_template_type="%s", ' \
                  'in_hierarchical=%s)' % (dn,
                                           sp_new_name,
                                           template_type,
                                           in_hierarchical_value)
        cmdlet += '\nmo_list = handle.process_xml_elem(elem)\n'

    return cmdlet


def _generate_multiple_clone_cmdlets(xml_string, is_prefix_based):
    """
    Internal method which takes xmlstring, and generate script for
    lsInstantiateNTemplate and  lsInstantiateNNamedTemplate methods.
    """

    doc = xml.dom.minidom.parseString(xml_string)
    node = None
    top_node = doc.documentElement
    if top_node is None:
        return

    if is_prefix_based:
        if top_node.localName == "lsInstantiateNTemplate":
            node = top_node
    else:
        if top_node.localName == "lsInstantiateNNamedTemplate":
            node = top_node

    dn = ""
    if node.hasAttribute(NamingPropertyId.DN):
        dn = node.getAttribute(NamingPropertyId.DN)
    else:
        print("Attribute 'Dn' not available")  # writewarning in dotnet
        return

    dest_org = ""
    if node.hasAttribute("inTargetOrg"):
        dest_org = node.getAttribute("inTargetOrg")

    sp_name = re.sub(
        r"^(org-[\-\.:_a-zA-Z0-9]{1,16}/)*org-[\-\.:_a-zA-Z0-9]{1,16}/ls-", "",
        dn)

    # source_org = ""
    match_object = re.match(
        r"^(org-[\-\.:_a-zA-Z0-9]{1,16}/)*org-[\-\.:_a-zA-Z0-9]{1,16}", dn)
    if match_object is not None:
        # source_org = match_object.group(0)
        pass
    else:
        print("Attribute 'Dn' is corrupt")
        return

    cmdlet = ""
    in_hierarchical = ""
    if node.hasAttribute("in_hierarchical") is not None:
        in_hierarchical = node.getAttribute("in_hierarchical")

    in_error_on_existing = ""
    if node.hasAttribute("inErrorOnExisting") is not None:
        in_error_on_existing = node.getAttribute("inErrorOnExisting")

    if is_prefix_based:
        # prefix = ""
        if node.hasAttribute("inServerNamePrefixOrEmpty") is not None:
            # prefix = node.getAttribute("inServerNamePrefixOrEmpty")
            pass
        else:
            print("Attribute 'inServerNamePrefixOrEmpty' not available")
            return

        count = 0
        if node.hasAttribute("inNumberOf") is not None:
            count = node.getAttribute("inNumberOf")
        else:
            print("Attribute 'inNumberOf' not available")
            return

        if in_hierarchical.lower() == "true":
            in_hierarchical_value = YesOrNo.TRUE
        else:
            in_hierarchical_value = YesOrNo.FALSE

        cmdlet = '\nfrom ucsmsdk.ucsmethodfactory import ' \
                 'ls_instantiate_n_template\n'
        cmdlet += 'elem = ls_instantiate_n_template(' \
                  'cookie=handle.cookie, dn=\%s", in_number_of=%s, ' \
                  'in_server_name_prefix_or_empty="%s", in_target_org="%s", ' \
                  'in_hierarchical=%s)' % (dn, count, sp_name, dest_org,
                                           in_hierarchical_value)
        cmdlet += '\nmo_list = handle.process_xml_elem(elem)\n'
    else:
        dn_nodes = _get_elem_child_nodes(
            node.getElementsByTagName("inNameSet")[0])
        if dn_nodes is None or len(dn_nodes) < 1:
            print("Xml is corrupt. New names not available")
            return

        new_names = "@("
        new_name_exists = False
        temp_dn = ""
        cmdlet = '\nfrom ucsmsdk.ucsmethodfactory import ' \
                 'ls_instantiate_n_named_template\n'
        cmdlet += 'from ucsmsdk.ucsbasetype import DnSet, Dn\n\n'
        cmdlet += "dn_set = DnSet()" + "\n"

        for dn_node in dn_nodes:
            if dn_node.hasAttribute("value"):
                new_name_exists = True
                temp_dn = dn_node.getAttribute("value")
                new_names += "\"" + temp_dn + "\","
                cmdlet += "dn = Dn()\ndn.attr_set(\"value\"," \
                          "\"%s\")\ndn_set.child_add(dn)\n" % temp_dn
            else:
                print("Xml is corrupt. New names not available")
                return

        if not new_name_exists:
            print("Xml is corrupt. New names not available")
            return

        new_names = new_names.rstrip(',')
        new_names += ")"

        if in_hierarchical.lower() == "true":
            in_hierarchical_value = YesOrNo.TRUE
        else:
            in_hierarchical_value = YesOrNo.FALSE

        cmdlet += 'elem = ls_instantiate_n_named_template(' \
                  'cookie=handle.cookie, dn="%s", in_error_on_existing=' \
                  '"%s", in_name_set=dn_set, in_target_org' \
                  '="%s", in_hierarchical="%s")' % (dn,
                                                    in_error_on_existing,
                                                    dest_org,
                                                    in_hierarchical_value)
        cmdlet += '\nmo_list = handle.process_xml_elem(elem)\n'
    return cmdlet


def _generate_clear_interval_cmdlet(xml_string):
    """
    Internal method which takes xmlstring, and generate script for
    statsClearInterval method.
    """

    doc = xml.dom.minidom.parseString(xml_string)
    node = None
    top_node = doc.documentElement
    if top_node is None:
        return

    if top_node.localName == "statsClearInterval":
        node = top_node
    else:
        print("Check if Method is <statsClearInterval>")
        return

    cmdlet = ""
    dn_nodes = _get_elem_child_nodes(node.getElementsByTagName("inDns")[0])

    if dn_nodes is None or len(dn_nodes) < 0:
        return

    cmdlet = "dn_set = DnSet()" + "\n"
    temp_dn = ""

    for dn_node in dn_nodes:
        temp_dn = dn_node.getAttribute(NamingPropertyId.VALUE)
        cmdlet += "dn = Dn()\ndn.attr_set(\"value\", " \
                  "\"%s\")\ndn_set.child_add(dn)\n" % temp_dn
    cmdlet += "handle.config_resolve_dns(dn_set)"

    return cmdlet


def _generate_config_conf_rename_cmdlet(xml_string):
    """
    Internal method which takes xmlstring, and generate script for
    configConfRename method.
    """

    doc = xml.dom.minidom.parseString(xml_string)
    node = None
    top_node = doc.documentElement
    if top_node is None:
        return

    if top_node.localName == "configConfRename":
        node = top_node
    else:
        print("Check if Method is <configConfRename>")
        return

    if node.hasAttribute('dn'):
        dn = node.getAttribute('dn')
    if node.hasAttribute('inNewName'):
        in_new_name = node.getAttribute('inNewName')
    if node.hasAttribute('inHierarchical'):
        in_hierarchical = node.getAttribute('inHierarchical')
    cmdlet = "\nfrom ucsmsdk.ucsmethodfactory import config_conf_rename\n\n"
    cmdlet += 'elem = config_conf_rename(' \
              'cookie=handle.cookie, dn="%s", ' \
              'in_new_name="%s", ' \
              'in_hierarchical="%s")' % (dn, in_new_name, in_hierarchical)
    cmdlet += '\nmo_list = handle.process_xml_elem(elem)\n'
    return cmdlet


# ---- END - OF - METHOD SPECIFIC - FUNCTION ----

def _generate_cmdlets(xml_string):
    """
    Internal method which takes xmlstring, and identify which method to call
    for a specific method.
    """

    doc = xml.dom.minidom.parseString(xml_string)
    cmdlet = ""
    global _display_xml, _outfile_path
    if _display_xml:
        # doc = xml.dom.minidom.parseString(xml_string)
        _dump_xml_on_screen(doc)

    category = ""
    match_found = re.match(r"^[\s]*<[\s]*([\S]+)", xml_string)
    if match_found:
        method_name = match_found.group(1)
        category = method_name
    else:
        return

    if category == "configConfMos" or category == "configConfMo":
        cmdlet = _generate_config_conf_cmdlets(xml_string)
    elif category in ["configResolveDn", "configResolveDns",
                      "configResolveClass", "configResolveClasses"]:
        cmdlet = _generate_config_resolve_cmdlet(xml_string, category)
    elif category == "lsClone":
        cmdlet = _generate_single_clone_cmdlets(xml_string, False)
    elif category == "lsInstantiateTemplate":
        cmdlet = _generate_single_clone_cmdlets(xml_string, True)
    elif category == "lsTemplatise":
        cmdlet = _generate_ls_templatise_cmdlets(xml_string)
    elif category == "lsInstantiateNTemplate":
        cmdlet = _generate_multiple_clone_cmdlets(xml_string, True)
    elif category == "lsInstantiateNNamedTemplate":
        cmdlet = _generate_multiple_clone_cmdlets(xml_string, False)
    elif category == "statsClearInterval":
        cmdlet = _generate_clear_interval_cmdlet(xml_string)
    elif category == "configConfRename":
        cmdlet = _generate_config_conf_rename_cmdlet(xml_string)
    # support for redirecting script output to respective file
    if _outfile_flag:
        _outfile = open(_outfile_path, 'a')
        _outfile.write('##### Start-Of-PythonScript #####')
        _outfile.write(cmdlet)
        _outfile.write('\n##### End-Of-PythonScript #####\n\n')
        _outfile.close()
    else:
        print("##### Start-Of-PythonScript #####")
        print(cmdlet)
        print("##### End-Of-PythonScript #####")
    return


def _extract_xml(file_stream, line):
    """
    Internal method which extracts xmlstring from the file stream and calls the
    _generate_cmdlets() on this xmlstring.
    """

    read_flag = False
    request_string = ""
    line_count = 0
    while line != "":
        if read_flag and not re.search(r"^\s*$", line):
            request_string += line + "\n"
            line_count += 1
        if _check_if_any_list_value_in_string(_multi_line_method, line):
            request_string += line + "\n"
            read_flag = True
            line_count += 1
        if read_flag and _check_if_any_list_value_in_string(
                _single_line_method, line):
            if line_count > 1 and '/>' in line:
                line = file_stream.readline()
                continue
            read_flag = False
            _generate_cmdlets(request_string)
            request_string = ""
            break

        line = file_stream.readline()


def _find_xml_requests_in_file(file_stream, gui_log):
    """
    Internal method which depending on gui_log flag, calls the _extract_xml()
    internally.
    """

    line = file_stream.readline()
    while line != "":
        if not gui_log:
            _extract_xml(file_stream, line)
        elif "[------------- Sending Request to Server ------------" in line:
            line = file_stream.readline()
            if line is not None and _check_if_any_list_value_in_string(
                    _multi_line_method,line):
                # print "[%s]" %(line)
                _extract_xml(file_stream, line)
        line = file_stream.readline()


def _if_path_or_literal_path(path, literal_path, gui_log):
    """
    Internal method which checks if path or literal_path present for the
    respective parameter set and if exists then
    call _find_xml_requests_in_file().
    """

    if path:
        if literal_path:
            print("Parameter <path> takes precedence over <literal_path>")
        file_path = path
    elif literal_path:
        file_path = literal_path
    else:
        print("Please provide path or literal_path")
        return

    file_stream = open(file_path, 'r')
    _find_xml_requests_in_file(file_stream, gui_log)
    file_stream.close()


def _get_ucs_default_logpath_windows():
    """
    Internal method to find ucsm logs path on windows os.
    """

    if 'APPDATA' in os.environ.keys():
        log_file_path = os.getenv('APPDATA')
    else:
        print(os.name)
        raise Exception('Not windows OS')

    if sys.getwindowsversion()[0] == 6:  # in case OS is Win 2008 or above
        log_file_path = os.path.join(dirname(log_file_path), "LocalLow")
    log_file_path += r"\Sun\Java\Deployment\log\.ucsm" + "\\"
    return log_file_path


def _get_ucs_default_logpath_linux():
    """
    Internal method to find ucsm logs path on linux os.
    """

    log_file_path = os.getenv('HOME')
    log_file_path += r"/.java/deployment/log/.ucsm/"
    return log_file_path


def _get_ucs_default_logpath_osx():
    """
    Internal method to find ucsm logs path on osx os.
    """

    log_file_path = os.getenv('HOME')
    # log_file_path += r"/Library/Caches/Java/log/.ucsm"
    path = r"/Library/Application Support/Oracle/Java/Deployment/log/.ucsm"
    log_file_path += path
    return log_file_path


def _get_ucs_default_logpath_cygwin():
    """
    Internal method to find ucsm logs path on cygwin os.
    """

    log_file_path = os.getenv('APPDATA')
    log_file_path = log_file_path.replace("\\", "/")
    # TODO:
    # if OSMajorVersion == 6:
    log_file_path = dirname(log_file_path) + "/LocalLow"
    log_file_path += r"/Sun/Java/Deployment/log/.ucsm/"
    return log_file_path


def convert_to_ucs_python(log_path=None, xml=False, request=None,
                          gui_log=False, path=None, literal_path=None,
                          dump_to_file=False, dump_file_path=None,
                          dump_xml=False):
    """
    This operation generates python script from ucsm gui using ucsm logs
    or directly from ucs logfile or xml request.

    Args:
        log_path (str): ucsm guilog path, if code unable to find the ucsm
              guilog path you can specify the path using this parameter.
        xml (bool): if True, generates python script using xml request.
        request (str): xml request
        gui_log (bool): if True, generates python script using ucsm logfile
        path (str): path of the file containing xml request or ucsm logfile
        literal_path (str): path of the file containing xml request or
              ucsm logfile
        dump_to_file (bool): if True, dump the code to file
        dump_file_path (str): path of file where to dump generated code.
        dump_xml (bool): if True, display the xml extracted for generating
              code

    Examples:
        * Optional in all param set - dump_to_file, dump_file_path, dump_xml

        * param set 1:
            convert_to_ucs_python()\n
            convert_to_ucs_python(dump_xml=True)\n
            file_path = r"/home/user/pythoncode.py"\n
            convert_to_ucs_python(dump_to_file=True,
                              dump_file_path=file_path,
                              dump_xml=True)\n

        * param set 2:
            xml_str='''
             <configConfRename
             dn="org-root/ls-ra1"
             inNewName="ra2"
             inHierarchical="false">
             </configConfRename>'''\n
            convert_to_ucs_python(xml=True, request=xml_str)\n

            file_path = '/home/user/ucsmxml/configrequest.xml'\n
            convert_to_ucs_python(xml=True, path=file_path)\n

        * param set 3:
            file_path = '/home/user/ucsmlog/centrale_14804.log'\n
            convert_to_ucs_python(gui_log=True, path=file_path)\n
    """

    print("### Please review the generated cmdlets before deployment.\n")
    global _display_xml, _outfile_flag, _outfile_path, _outfile
    _display_xml = dump_xml
    _outfile_flag = dump_to_file
    _outfile_path = dump_file_path

    if _outfile_flag in ucsgenutils.AFFIRMATIVE_LIST:
        if _outfile_path:
            print("### Script Output is in file < " + _outfile_path + " >")
            _outfile = open(_outfile_path, 'w')
            _outfile.close()
            # _outfile = open(r"c:\work.txt", 'w+')
        else:
            print("Provide dump_file_path")
            return
    if xml in ucsgenutils.AFFIRMATIVE_LIST:
        if gui_log in ucsgenutils.AFFIRMATIVE_LIST:
            print("parameter <xml> takes precedence over <gui_log>")
        if request:
            _generate_cmdlets(request)
        elif path or literal_path:
            _if_path_or_literal_path(path, literal_path, False)
        else:
            print("Provide request")
            return
    elif gui_log in ucsgenutils.AFFIRMATIVE_LIST:
        if path or literal_path:
            _if_path_or_literal_path(path, literal_path, True)
        else:
            print("Provide path or literal_path")
    else:
        if log_path:
            if not os.path.exists(log_path):
                raise ValueError("%s does not exists." % log_path)
            if not os.path.isdir(log_path):
                raise ValueError("%s is not a directory." % log_path)
            log_file_path = log_path
        else:
            from sys import platform as _platform

            if _platform.lower() == "linux" or _platform.lower() == "linux2":
                # linux
                log_file_path = _get_ucs_default_logpath_linux()
            elif _platform.lower() == "darwin":
                # OS X
                log_file_path = _get_ucs_default_logpath_osx()
            elif _platform.lower() == "win32" or _platform.lower() == "win64":
                # Windows...
                log_file_path = _get_ucs_default_logpath_windows()
            elif "cygwin" in _platform.lower():
                # Cygwin
                log_file_path = _get_ucs_default_logpath_cygwin()
            else:
                print("[Error]: Unsupported OS: %s" % _platform)
                log_file_path = None
                return

        os.chdir(log_file_path)
        files = [file_ for file_ in glob.glob("centrale_*.log") if
                 os.path.isfile(file_)]
        files.sort(key=lambda x: os.path.getmtime(x), reverse=True)
        last_updated_file = files[0]

        print("ucsm logfile: %s\n" % (
            os.path.join(log_file_path, last_updated_file)))

        file_stream = open(last_updated_file, 'r')
        # read the file till the end
        cnt = 0
        for line in file_stream:
            cnt += 1
        # Wait indefinitely until receive new line of set and then again wait
        while True:
            # line = file_stream.readline()
            # if line:
            #     _find_xml_requests_in_file_test(file_stream, True)
            _find_xml_requests_in_file(file_stream, True)
            time.sleep(2)
        file_stream.close()

        # if _outfile_path:
        # _outfile.close()

    print("### End of Convert-To-Python ###")



# Copyright 2013 Cisco Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the APIs used to create and download tech_support file.
"""

import os
import time
import datetime
import logging
from ..deprecated import deprecated
from ..ucsexception import UcsValidationException, UcsWarning
from ..mometa.sysdebug.SysdebugTechSupportCmdOpt import \
    SysdebugTechSupportCmdOptConsts as TechSupOptsConsts

log = logging.getLogger('ucs')


def _create_download_dir(file_dir):
    if not os.path.exists(file_dir):
        os.makedirs(file_dir)


def _validate_download_args(file_dir, file_name):
    if file_dir is None:
        raise UcsValidationException('file_dir is a required parameter')

    if file_name is None:
        raise UcsValidationException('file_name is a required parameter')

    if not file_name.endswith('.tar'):
        raise UcsValidationException('file_name should end with .tar')


def _get_creation_ts():
    dt1 = datetime.datetime(1970, 1, 1, 12, 0, 0, 0)
    dt2 = datetime.datetime.utcnow()
    return int((dt2 - dt1).total_seconds())


def _bootstrap_tech_support_mo(creation_ts):
    from ..mometa.top.TopSystem import TopSystem
    from ..mometa.sysdebug.SysdebugTechSupport import SysdebugTechSupport, \
        SysdebugTechSupportConsts
    from ..mometa.sysdebug.SysdebugTechSupFileRepository import \
        SysdebugTechSupFileRepository

    top_system = TopSystem()
    ts_file_repo = SysdebugTechSupFileRepository(parent_mo_or_dn=top_system)
    ts_mo = SysdebugTechSupport(
        parent_mo_or_dn=ts_file_repo,
        creation_ts=str(creation_ts),
        admin_state=SysdebugTechSupportConsts.ADMIN_STATE_START)
    return ts_mo


def _is_valid_arg(param, kwargs):
    return param in kwargs and kwargs[param] is not None


def _set_ts_options_ucsm(ts_cmd_opt, kwargs):
    ts_cmd_opt.major_opt_type = TechSupOptsConsts.MAJOR_OPT_TYPE_UCSM


def _set_ts_options_ucsm_mgmt(ts_cmd_opt, kwargs):
    ts_cmd_opt.major_opt_type = TechSupOptsConsts.MAJOR_OPT_TYPE_UCSM_MGMT


def _validate_chassis_options(kwargs):
    if not _is_valid_arg("chassis_id", kwargs):
        raise UcsValidationException('chassis_id should be specified')
    if "cimc_id" not in kwargs and "iom_id" not in kwargs and "cartridge_id" not in kwargs:
        raise UcsValidationException('cimc_id/iom_id/cartridge_id should be specified')


def _set_ts_options_chassis(ts_cmd_opt, kwargs):
    _validate_chassis_options(kwargs)
    ts_cmd_opt.chassis_id = str(kwargs["chassis_id"])
    ts_cmd_opt.major_opt_type = TechSupOptsConsts.MAJOR_OPT_TYPE_CHASSIS
    if _is_valid_arg("cimc_id", kwargs):
        ts_cmd_opt.chassis_cimc_id = str(kwargs["cimc_id"])

        if _is_valid_arg("adapter_id", kwargs):
            ts_cmd_opt.cimc_adapter_id = str(kwargs["adapter_id"])
        else:
            ts_cmd_opt.cimc_adapter_id = TechSupOptsConsts.CIMC_ADAPTER_ID_ALL
    elif _is_valid_arg("iom_id", kwargs):
        ts_cmd_opt.chassis_iom_id = str(kwargs["iom_id"])
    elif _is_valid_arg("cartridge_id", kwargs):
        ts_cmd_opt.chassis_cartridge_id = str(kwargs["cartridge_id"])
        if ts_cmd_opt.chassis_cartridge_id != "all":
            if _is_valid_arg("cartridge_cimc_id", kwargs):
                ts_cmd_opt.cartridge_cimc_id = str(kwargs["cartridge_cimc_id"])
            else:
                raise UcsValidationException('cartridge_cimc_id should be specified')


def _validate_rackserver_options(kwargs):
    if not _is_valid_arg("rack_server_id", kwargs):
        raise UcsValidationException('rack_server_id should be specified')


def _set_ts_options_rackserver(ts_cmd_opt, kwargs):
    _validate_rackserver_options(kwargs)
    ts_cmd_opt.rack_server_id = str(kwargs["rack_server_id"])
    ts_cmd_opt.major_opt_type = TechSupOptsConsts.MAJOR_OPT_TYPE_SERVER

    if _is_valid_arg("rack_adapter_id", kwargs):
        ts_cmd_opt.rack_server_adapter_id = str(kwargs["rack_adapter_id"])
    else:
        ts_cmd_opt.rack_server_adapter_id = TechSupOptsConsts.RACK_SERVER_ADAPTER_ID_ALL


def _validate_fex_options(kwargs):
    if not _is_valid_arg("fex_id", kwargs):
        raise UcsValidationException('fex_id should be specified')


def _set_ts_options_fex(ts_cmd_opt, kwargs):
    _validate_fex_options(kwargs)
    ts_cmd_opt.fab_ext_id = str(kwargs["fex_id"])
    ts_cmd_opt.major_opt_type = TechSupOptsConsts.MAJOR_OPT_TYPE_FEX


def _validate_servermemory_options(kwargs):
    if not _is_valid_arg("server_id_list", kwargs):
        raise UcsValidationException('server_id_list should be specified')


def _set_ts_options_servermemory(ts_cmd_opt, kwargs):
    _validate_servermemory_options(kwargs)
    ts_cmd_opt.server_id_list = str(kwargs["server_id_list"])
    ts_cmd_opt.major_opt_type = TechSupOptsConsts.MAJOR_OPT_TYPE_SERVER_MEMORY


def _set_ts_command_options(ts_cmd_opt, kwargs):
    if _is_valid_arg("command_options", kwargs):
        ts_cmd_opt.command_options = str(kwargs["command_options"])


def _set_ts_options(option, ts_cmd_opt, kwargs):
    if option == "ucsm":
        _set_ts_options_ucsm(ts_cmd_opt, kwargs)
    elif option == "ucsm-mgmt":
        _set_ts_options_ucsm_mgmt(ts_cmd_opt, kwargs)
    elif option == "chassis":
        _set_ts_options_chassis(ts_cmd_opt, kwargs)
    elif option == "fabric-extender":
        _set_ts_options_fex(ts_cmd_opt, kwargs)
    elif option == "rack-server":
        _set_ts_options_rackserver(ts_cmd_opt, kwargs)
    elif option == "server-memory":
        _set_ts_options_servermemory(ts_cmd_opt, kwargs)
    else:
        raise UcsValidationException('Unrecognised option value: ' + option)

    _set_ts_command_options(ts_cmd_opt, kwargs)


def _check_for_failure(err):
    # The following error messages define the criteria for
    # the poll loop to be interrupted by an exception.
    # There are occasinal Request Timeouts, which could be ignored
    # if the subsequent calls are being replied to by the server.
    failure_conditions = ["urlopen error",
                          "TechSupport creation timed out",
                          "TechSupport creation failed"]
    for each in failure_conditions:
        if each in str(err):
            raise Exception(each)


def _fail_and_remove_ts(handle, ts_mo, err):
    if ts_mo:
        handle.remove_mo(ts_mo)
        handle.commit()
    raise UcsValidationException(err)


def poll_wait_for_tech_support(handle, ts_mo, timeout):
    from ..mometa.sysdebug.SysdebugTechSupport import \
        SysdebugTechSupportConsts as TechSupConsts

    poll_interval = 5
    duration = timeout
    while True:
        try:
            ts = handle.query_dn(ts_mo.dn)
            if ts is None:
                _fail_and_remove_ts(handle, ts, 'TechSupport creation failed')
            if ts.oper_state == TechSupConsts.OPER_STATE_AVAILABLE:
                break
            if ts.oper_state == TechSupConsts.OPER_STATE_FAILED or \
                    ts.oper_state == TechSupConsts.OPER_STATE_UNAVAILABLE:
                _fail_and_remove_ts(handle, ts, 'TechSupport creation failed')
            time.sleep(min(duration, poll_interval))
            duration = max(0, (duration - poll_interval))
            if duration == 0:
                _fail_and_remove_ts(
                    handle, ts, 'TechSupport creation timed out')
        except Exception as err:
            _check_for_failure(err)


def download_tech_support(handle, name, file_dir, file_name):
    try:
        handle.file_download(url_suffix="techsupport/" + name,
                             file_dir=str(file_dir),
                             file_name=str(file_name))
    except Exception as err:
        UcsWarning(str(err))


def remove_tech_support(handle, ts_mo):
    ts_mo.admin_state = "delete"
    handle.set_mo(ts_mo)
    handle.commit()


def get_tech_support(handle,
                     option="ucsm",
                     remove_from_ucs=False,
                     download=True, file_dir=".", file_name="techsupport.tar",
                     timeout=1200,
                     **kwargs):
    """
    Create and optionally download technical support for Ucs or individual
        components.

    Args:
        handle (UcsHandle): Ucs connection handle
        option (str): categorty of technical support to be created.
                      possible values:
                        - ucsm
                        - ucsm-mgmt
                        - chassis:
                            Mandates that user specifies "chassis_id"
                            Either "cimc_id" or "iom_id" or "cartridge_id"
                            When specific "cartridge_id" is given, instead of "all",
                            mandatory "cartridge_cimc_id" must also be given
                        - fabric-extender
                            Mandates that user specifies "fex_id"
                        - rack-server
                            Mandates that user specifies "rack_server_id"
                        - server-memory
                            Mandates that user specifies "server_id_list"
        remove_from_ucs (bool): Remove the created technical support, if True
        download (bool): download the technical support if True
        kwargs: key=value pairs relevant to the seleced option

    Examples:
        get_tech_support(handle,
                         option="ucsm",
                         file_dir=".",
                         file_name="techsupport.tar")

        get_tech_support(handle,
                         option="ucsm-mgmt",
                         file_dir=".",
                         file_name="techsupport.tar")

        get_tech_support(handle,
                         option="chassis",
                         file_dir=".",
                         file_name="techsupport.tar",
                         chassis_id=1,
                         cimc_id=1,
                         adapter_id=1)

        get_tech_support(handle,
                         option="chassis",
                         file_dir=".",
                         file_name="techsupport.tar",
                         chassis_id=1,
                         iom_id=1)

        get_tech_support(handle,
                         option="chassis",
                         file_dir=".",
                         file_name="techsupport.tar",
                         chassis_id=1,
                         cartridge_id=1,
                         cartridge_cimc_id=1)

        get_tech_support(handle,
                         option="rack-server",
                         file_dir=".",
                         file_name="techsupport.tar",
                         rack_server_id=1,
                         rack_adapter_id="all")

        get_tech_support(handle,
                         option="fabric-extender",
                         file_dir=".",
                         file_name="techsupport.tar",
                         fex_id=1)

        get_tech_support(handle,
                         option="server-memory",
                         file_dir=".",
                         file_name="techsupport.tar",
                         server_id_list="1,2")

    """
    from ..mometa.sysdebug.SysdebugTechSupportCmdOpt import \
        SysdebugTechSupportCmdOpt

    if download:
        _validate_download_args(file_dir, file_name)
        _create_download_dir(file_dir)

    ts_mo = _bootstrap_tech_support_mo(_get_creation_ts())
    ts_cmd_opt = SysdebugTechSupportCmdOpt(parent_mo_or_dn=ts_mo)
    _set_ts_options(option, ts_cmd_opt, kwargs)

    handle.add_mo(ts_mo)
    handle.commit()

    poll_wait_for_tech_support(handle, ts_mo, timeout)

    # The backend changes the name of the MO in some scenarios.
    # Query it again to have the updated MO name
    ts_mo = handle.query_dn(ts_mo.dn)

    if download:
        download_tech_support(handle, ts_mo.name, file_dir, file_name)

    if remove_from_ucs:
        remove_tech_support(handle, ts_mo)


@deprecated(get_tech_support)
def get_ucs_tech_support(handle,
                         ucs_manager=False,
                         ucs_mgmt=False,
                         chassis_id=None, cimc_id=None,
                         adapter_id=None, iom_id=None,
                         fex_id=None,
                         rack_server_id=None, rack_adapter_id=None,
                         remove_from_ucs=False,
                         download_techsupp=True, file_dir=None, file_name=None,
                         timeout_in_sec=600):
    """
    This operation creates and downloads the technical support file for
    the specified Ucs server.

    Args:
        handle (UcsHandle): Ucs connection handle
        ucs_manager (bool): True/False,
                            False - by default
                            Create and download TechSupport for UCSM, if true
        ucs_mgmt (bool): True/False,
                         False - by default
                         Create and download TechSupport for UCSM Management
                         services(excluding Fabric interconnects), if true
        chassis_id (int): chassis id
        cimc_id (int/string): for a specific chassis. Can be 'all'.
        adapter_id (int/string): for a specific chassis. Can be 'all'.
        iom_id (int/string): for a specific chassis. Can be 'all'.
        fex_id (int): id of a fabric extender.
        rack_server_id (int): id of a rack server.
        rack_adapter_id (int/string): adaptor_id for a specific rack server.
                              Can be 'all'.
        remove_from_ucs (bool): True/False,
                                False - by default
                                TechSupport will be removed from server, if True
        download_techsupp (bool): True/False,
                                    True - by default
                                    Download the TechSupport file, if True
        file_dir (str): directory to download tech support file to
        file_name (str): name of the download tech support file
        timeout_in_sec (int): specifies the time in seconds after which
                                the operation times-out.

    Example:
        * M - Manadatory, O - Optional
        * Note:
          Mandatory in ALL param sets: file_dir, file_name
          Optional in ALL param sets: timeout_in_sec, remove_from_ucs,
                                      download_techsupp

        * param set 1:
        * M - ucs_manager
        ---------------------------------------------------------------
        file_dir = "/home/user/techsupp"
        file_name = "techsupp_ucs_manager.tar"
        get_ucs_tech_support(handle,
                            file_dir=file_dir,
                            file_name=file_name,
                            ucs_manager=True)

        get_ucs_tech_support(handle,
                    file_dir=file_dir,
                    file_name=file_name,
                    ucs_manager=True,
                    timeout_in_sec=300,
                    remove_from_ucs=True)

        * param set 2:
        * M - ucs_manager
        ---------------------------------------------------------------
        file_dir = "/home/user/techsupp"
        file_name = "techsupp_ucs_mgmt.tar"
        get_ucs_tech_support(handle,
                            file_dir=file_dir,
                            file_name=file_name,
                            ucs_mgmt=True)

        * param set 3:
        * M - chassis_id, cimc_id
        * O - adapter_id
        ---------------------------------------------------------------
        file_dir = "/home/user/techsupp"
        file_name = "techsupp_ucs_mgmt.tar"
        get_ucs_tech_support(handle,
                            file_dir=file_dir,
                            file_name=file_name,
                            chassis_id=1,
                            cimc_id=1,
                            adapter_id=1)

        * param set 4:
        * M - chassis_id, iom_id
        ---------------------------------------------------------------
        file_dir = "/home/user/techsupp"
        file_name = "techsupp_ucs_mgmt.tar"
        get_ucs_tech_support(handle,
                            file_dir=file_dir,
                            file_name=file_name,
                            chassis_id=1,
                            iom_id=1)

        * param set 5:
        * M - fex_id
        ---------------------------------------------------------------
        file_dir = "/home/user/techsupp"
        file_name = "techsupp_ucs_mgmt.tar"
        get_ucs_tech_support(handle,
                            file_dir=file_dir,
                            file_name=file_name,
                            fex_id=1)

        * param set 6:
        * M - rack_server_id
        * O - rack_adapter_id
        ---------------------------------------------------------------
        file_dir = "/home/user/techsupp"
        file_name = "techsupp_ucs_mgmt.tar"
        get_ucs_tech_support(handle,
                            file_dir=file_dir,
                            file_name=file_name,
                            rack_server_id=1,
                            rack_adapter_id=1)
    """

    from ..mometa.top.TopSystem import TopSystem
    from ..mometa.sysdebug.SysdebugTechSupport import SysdebugTechSupport, \
        SysdebugTechSupportConsts
    from ..mometa.sysdebug.SysdebugTechSupFileRepository import \
        SysdebugTechSupFileRepository
    from ..mometa.sysdebug.SysdebugTechSupportCmdOpt import \
        SysdebugTechSupportCmdOpt, SysdebugTechSupportCmdOptConsts

    if download_techsupp:
        if file_name is None:
            raise UcsValidationException('provide file_name')
        if file_dir is None:
            raise UcsValidationException('provide dir_name')

        if not file_name.endswith('.tar'):
            raise UcsValidationException('file_name should end with .tar')

        if not os.path.exists(file_dir):
            os.makedirs(file_dir)

    # Converting timedelta in to total seconds for Python version compatibility
    dt1 = datetime.datetime(1970, 1, 1, 12, 0, 0, 0)
    dt2 = datetime.datetime.utcnow()
    creation_ts = int((dt2 - dt1).total_seconds())

    # create SysdebugTechSupport
    top_system = TopSystem()
    sysdebug_techsup_file_repo = SysdebugTechSupFileRepository(
        parent_mo_or_dn=top_system)
    sys_debug_tech_support = SysdebugTechSupport(
        parent_mo_or_dn=sysdebug_techsup_file_repo,
        creation_ts=str(creation_ts),
        admin_state=SysdebugTechSupportConsts.ADMIN_STATE_START)

    sys_debug_tech_support_cmd_opt = SysdebugTechSupportCmdOpt(
        parent_mo_or_dn=sys_debug_tech_support)

    # Parameter Set UCSM
    if ucs_manager:
        sys_debug_tech_support_cmd_opt.major_opt_type = \
            SysdebugTechSupportCmdOptConsts.MAJOR_OPT_TYPE_UCSM
    elif ucs_mgmt:
        sys_debug_tech_support_cmd_opt.major_opt_type = \
            SysdebugTechSupportCmdOptConsts.MAJOR_OPT_TYPE_UCSM_MGMT
    elif chassis_id is not None:
        if cimc_id is not None:
            sys_debug_tech_support_cmd_opt.chassis_cimc_id = str(cimc_id)
            sys_debug_tech_support_cmd_opt.chassis_id = str(chassis_id)
            sys_debug_tech_support_cmd_opt.major_opt_type = \
                SysdebugTechSupportCmdOptConsts.MAJOR_OPT_TYPE_CHASSIS

            if adapter_id is None:
                sys_debug_tech_support_cmd_opt.cimc_adapter_id = \
                    SysdebugTechSupportCmdOptConsts.CIMC_ADAPTER_ID_ALL
            else:
                sys_debug_tech_support_cmd_opt.cimc_adapter_id = \
                    str(adapter_id)
        elif iom_id is not None:
            sys_debug_tech_support_cmd_opt.chassis_iom_id = str(iom_id)
            sys_debug_tech_support_cmd_opt.chassis_id = str(chassis_id)
            sys_debug_tech_support_cmd_opt.major_opt_type = \
                SysdebugTechSupportCmdOptConsts.MAJOR_OPT_TYPE_CHASSIS
    elif rack_server_id is not None:
        sys_debug_tech_support_cmd_opt.rack_server_id = str(rack_server_id)
        sys_debug_tech_support_cmd_opt.major_opt_type = \
            SysdebugTechSupportCmdOptConsts.MAJOR_OPT_TYPE_SERVER

        if rack_adapter_id is None:
            sys_debug_tech_support_cmd_opt.rack_server_adapter_id = \
                SysdebugTechSupportCmdOptConsts.RACK_SERVER_ADAPTER_ID_ALL
        else:
            sys_debug_tech_support_cmd_opt.rack_server_adapter_id = \
                str(rack_adapter_id)
    elif fex_id is not None:
        sys_debug_tech_support_cmd_opt.fab_ext_id = str(fex_id)
        sys_debug_tech_support_cmd_opt.major_opt_type = \
            SysdebugTechSupportCmdOptConsts.MAJOR_OPT_TYPE_FEX

    handle.add_mo(sys_debug_tech_support)
    handle.commit()

    # poll for tech support to complete
    duration = timeout_in_sec
    poll_interval = 2
    status = False

    while True:
        tech_support = handle.query_dn(sys_debug_tech_support.dn)
        if tech_support.oper_state == \
                SysdebugTechSupportConsts.OPER_STATE_AVAILABLE:
            status = True
        if status:
            break
        time.sleep(min(duration, poll_interval))
        duration = max(0, (duration - poll_interval))
        if duration == 0:
            handle.remove_mo(tech_support)
            handle.commit()
            raise UcsValidationException('TechSupport file creation timed out')

    # download tech support file
    if download_techsupp:
        url_suffix = "techsupport/" + tech_support.name
        try:
            handle.file_download(url_suffix=url_suffix,
                                 file_dir=file_dir,
                                 file_name=file_name)
        except Exception as err:
            UcsWarning(str(err))

    # remove tech support file from ucs
    if remove_from_ucs:
        tech_support.admin_state = "delete"
        handle.set_mo(tech_support)
        handle.commit()

    return tech_support



# Copyright 2013 Cisco Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the APIs used to launch ucs kvm.
"""
import subprocess
import logging

from .. import ucsgenutils
from ..ucsconstants import NamingId, YesOrNo
from ..ucsexception import UcsWarning, UcsValidationException, UcsException
from ..ucscoremeta import UcsVersion
from six.moves.urllib.parse import urlencode


log = logging.getLogger('ucs')


class _ParamKvm(object):
    # Internal class to act as enum to support ucs_kvm_launch utility.

    CENTRALE_PASSWORD = "centralePassword"
    CENTRALE_USER = "centraleUser"
    DN = "dn"
    FRAME_TITLE = "frame_title"
    KVM_IP_ADDR = "kvmIpAddr"
    KVM_PASSWORD = "kvmPassword"
    KVM_PN_DN = "kvmPnDn"
    KVM_USER = "kvmUser"
    TEMP_UNPW = "tempunpw"
    KVM_DN = "kvmDn"


def ucs_kvm_launch(handle, service_profile=None, blade=None, rack_unit=None,
                   frame_title=None, need_url=False):
    """
    ucs_kvm_launch launches the kvm session for a specified server.

    Args:
        handle (UcsHandle)
        service_profile (LsServer)
        blade (ComputeBlade)
        rack_unit (ComputeRackUnit)
        frame_title (str): title of launched frame
        need_url (bool): True/False,
                         Returns URL to launch kvm, if True

    Example:
        # sp is service profile object \n
        ucs_kvm_launch(handle, service_profile=sp)\n
        ucs_kvm_launch(handle, service_profile=sp, frame_title="using sp")\n
        ucs_kvm_launch(handle, service_profile=sp, need_url=True)\n

        # blade1 is ComputeBlade object\n
        ucs_kvm_launch(handle, blade=blade1)\n

        # rack1 is ComputeRackUnit object\n
        ucs_kvm_launch(handle, rack_unit=rack1)\n
    """

    from ..mometa.mgmt.MgmtIf import MgmtIfConsts
    from ..ucsmethodfactory import config_scope
    from ..ucsmethodfactory import aaa_get_n_compute_auth_token_by_dn

    if (blade is not None and rack_unit is not None) or \
        (service_profile is not None and rack_unit is not None) or \
            (blade is not None and service_profile is not None):
        raise UcsValidationException(
            "Provide only one parameter from blade, "
            "rack_unit and service profile.")
    if service_profile is None and blade is None and rack_unit is None:
        raise UcsValidationException(
            "Provide at least one parameter from blade, rack_unit and "
            "service profile.")

    min_version = UcsVersion('1.4(1a)')
    if handle.version < min_version:
        raise UcsValidationException(
            "start_kvm_session not supported for Ucs version older than %s. "
            "You are connected to Ucs Version %s" % (min_version,
                                                     handle.version))

    # lock = Lock()
    sp_bool = False
    nvc = {}
    dn = None
    pn_dn = None
    ip_address = None

    if blade is not None or rack_unit is not None:
        if blade is not None:
            pn_dn = blade.dn
        else:
            pn_dn = rack_unit.dn

        nvc[_ParamKvm.DN] = pn_dn

        if frame_title is None:
            frame_title = handle.name + ':' + pn_dn + ' KVM Console'
        nvc[_ParamKvm.FRAME_TITLE] = frame_title
        nvc[_ParamKvm.KVM_PN_DN] = pn_dn

        elem = config_scope(cookie=handle.cookie,
                            dn=pn_dn,
                            in_class=NamingId.MGMT_IF,
                            in_filter=None,
                            in_recursive=YesOrNo.FALSE,
                            in_hierarchical=YesOrNo.FALSE)

        response = handle.post_elem(elem)
        if response.error_code == 0:
            for mgmt_if in response.out_configs.child:
                if mgmt_if.subject == MgmtIfConsts.SUBJECT_BLADE and \
                   mgmt_if.admin_state == MgmtIfConsts.ADMIN_STATE_ENABLE:
                    ip_address = mgmt_if.ext_ip
        else:
            raise UcsException(response.error_code, response.error_descr)

        # If the blade does not have an IP,
        # check if a service profile is associated
        if ip_address is None or ip_address == '0.0.0.0':
            mo = handle.query_dn(dn=pn_dn)
            dn = mo.assigned_to_dn
            if dn is not None:
                sp_bool = True

    if sp_bool or service_profile is not None:
        if dn is None:
            dn = service_profile.dn
            if frame_title is None:
                frame_title = handle.name + ':' + dn + ' KVM Console'
            nvc[_ParamKvm.FRAME_TITLE] = frame_title
            sp_mo = service_profile
        else:
            sp_mo = handle.query_dn(dn)

        nvc[_ParamKvm.KVM_DN] = dn

        # sp_mo = service_profile

        if not sp_mo.pn_dn:
            raise UcsValidationException(
                'Service Profile is not associated with blade or rack_unit.')

        pn_dn = sp_mo.pn_dn
        nvc[_ParamKvm.DN] = pn_dn

        mo_list = handle.query_children(in_dn=dn, class_id='vnicIpV4Addr')
        # TODO:replace class_id with proper constantafter generating mos.py
        # or Constant.py

        for mo in mo_list:
            # gmo = _GenericMo(mo=mo, option=WriteXmlOption.ALL)
            # if 'Addr' not in gmo.properties:
            #     continue
            # ip_address = gmo.GetAttribute('Addr')
            # if ip_address is not None:
            #     break
            if hasattr(mo, "addr") or hasattr(mo, "Addr"):
                if hasattr(mo, "addr"):
                    ip_address = getattr(mo, "addr")
                elif hasattr(mo, "Addr"):
                    ip_address = getattr(mo, "Addr")
                if ip_address is not None:
                    break

    if (ip_address is None or ip_address == '0.0.0.0') and \
            service_profile is not None:

        elem = config_scope(cookie=handle.cookie,
                            dn=pn_dn,
                            in_class=NamingId.MGMT_IF,
                            in_filter=None,
                            in_recursive=YesOrNo.FALSE,
                            in_hierarchical=YesOrNo.FALSE)

        response = handle.post_elem(elem)
        if response.error_code == 0:
            for mgmt_if in response.out_configs.child:
                if mgmt_if.subject == MgmtIfConsts.SUBJECT_BLADE and \
                   mgmt_if.admin_state == MgmtIfConsts.ADMIN_STATE_ENABLE:
                    ip_address = mgmt_if.ext_ip

    if ip_address is None or ip_address == '0.0.0.0':
        raise UcsValidationException("No assigned IP address to use.")

    nvc[_ParamKvm.KVM_IP_ADDR] = ip_address
    elem = aaa_get_n_compute_auth_token_by_dn(cookie=handle.cookie,
                                              in_cookie=handle.cookie,
                                              in_dn=pn_dn,
                                              in_number_of=2)
    response = handle.post_elem(elem)
    if response.error_code == 0:
        nvc[_ParamKvm.CENTRALE_PASSWORD] = response.out_tokens.split(',')[0]
        nvc[_ParamKvm.CENTRALE_USER] = response.out_user
        nvc[_ParamKvm.KVM_PASSWORD] = response.out_tokens.split(',')[1]
        nvc[_ParamKvm.KVM_USER] = response.out_user
    else:
        raise UcsException(response.error_code, response.error_descr)

    nvc[_ParamKvm.TEMP_UNPW] = "true"
    uri = handle.uri
    kvm_url = '%s/ucsm/kvm.jnlp?%s' % (uri, urlencode(nvc))

    if need_url:
        return kvm_url
    else:
        install_path = ucsgenutils.get_java_installation_path()
        if install_path is not None:
            subprocess.call([install_path, kvm_url])
        else:
            UcsWarning("Java is not installed on System.")
            subprocess.Popen(kvm_url)



"""This module contains the general information for EquipmentEmbeddedStorageDef ManagedObject."""

from ...ucsmo import ManagedObject
from ...ucscoremeta import MoPropertyMeta, MoMeta
from ...ucsmeta import VersionMeta


class EquipmentEmbeddedStorageDefConsts:
    BLOCK_SIZE_512 = "512"
    BLOCK_SIZE_UNKNOWN = "unknown"
    CONNECTION_PROTOCOL_NVME = "NVME"
    CONNECTION_PROTOCOL_SAS = "SAS"
    CONNECTION_PROTOCOL_SATA = "SATA"
    CONNECTION_PROTOCOL_UNSPECIFIED = "unspecified"
    DEVICE_TYPE_HDD = "HDD"
    DEVICE_TYPE_SSD = "SSD"
    DEVICE_TYPE_UNSPECIFIED = "unspecified"
    INT_ID_NONE = "none"
    NUMBER_OF_BLOCKS_UNKNOWN = "unknown"
    POLICY_OWNER_LOCAL = "local"
    POLICY_OWNER_PENDING_POLICY = "pending-policy"
    POLICY_OWNER_POLICY = "policy"


class EquipmentEmbeddedStorageDef(ManagedObject):
    """This is EquipmentEmbeddedStorageDef class."""

    consts = EquipmentEmbeddedStorageDefConsts()
    naming_props = set([])

    mo_meta = MoMeta("EquipmentEmbeddedStorageDef", "equipmentEmbeddedStorageDef", "embedded-storage", VersionMeta.Version312b, "InputOutput", 0xff, [], [""], ['equipmentLocalDiskControllerCapProvider'], [], [None])

    prop_meta = {
        "block_size": MoPropertyMeta("block_size", "blockSize", "string", VersionMeta.Version312b, MoPropertyMeta.READ_ONLY, None, None, None, None, ["512", "unknown"], ["0-4294967295"]),
        "child_action": MoPropertyMeta("child_action", "childAction", "string", VersionMeta.Version312b, MoPropertyMeta.INTERNAL, 0x2, None, None, r"""((deleteAll|ignore|deleteNonPresent),){0,2}(deleteAll|ignore|deleteNonPresent){0,1}""", [], []),
        "connection_protocol": MoPropertyMeta("connection_protocol", "connectionProtocol", "string", VersionMeta.Version312b, MoPropertyMeta.READ_ONLY, None, None, None, None, ["NVME", "SAS", "SATA", "unspecified"], []),
        "descr": MoPropertyMeta("descr", "descr", "string", VersionMeta.Version312b, MoPropertyMeta.READ_WRITE, 0x4, None, None, r"""[ !#$%&\(\)\*\+,\-\./:;\?@\[\]_\{\|\}~a-zA-Z0-9]{0,256}""", [], []),
        "device_type": MoPropertyMeta("device_type", "deviceType", "string", VersionMeta.Version312b, MoPropertyMeta.READ_ONLY, None, None, None, None, ["HDD", "SSD", "unspecified"], []),
        "dn": MoPropertyMeta("dn", "dn", "string", VersionMeta.Version312b, MoPropertyMeta.READ_ONLY, 0x8, 0, 256, None, [], []),
        "int_id": MoPropertyMeta("int_id", "intId", "string", VersionMeta.Version312b, MoPropertyMeta.INTERNAL, None, None, None, None, ["none"], ["0-4294967295"]),
        "name": MoPropertyMeta("name", "name", "string", VersionMeta.Version312b, MoPropertyMeta.READ_WRITE, 0x10, None, None, r"""[\-\.:_a-zA-Z0-9]{0,16}""", [], []),
        "number_of_blocks": MoPropertyMeta("number_of_blocks", "numberOfBlocks", "string", VersionMeta.Version312b, MoPropertyMeta.READ_ONLY, None, None, None, None, ["unknown"], ["0-18446744073709551615"]),
        "policy_level": MoPropertyMeta("policy_level", "policyLevel", "uint", VersionMeta.Version312b, MoPropertyMeta.READ_ONLY, None, None, None, None, [], []),
        "policy_owner": MoPropertyMeta("policy_owner", "policyOwner", "string", VersionMeta.Version312b, MoPropertyMeta.READ_WRITE, 0x20, None, None, None, ["local", "pending-policy", "policy"], []),
        "rn": MoPropertyMeta("rn", "rn", "string", VersionMeta.Version312b, MoPropertyMeta.READ_ONLY, 0x40, 0, 256, None, [], []),
        "sacl": MoPropertyMeta("sacl", "sacl", "string", VersionMeta.Version312b, MoPropertyMeta.READ_ONLY, None, None, None, r"""((none|del|mod|addchild|cascade),){0,4}(none|del|mod|addchild|cascade){0,1}""", [], []),
        "size": MoPropertyMeta("size", "size", "ulong", VersionMeta.Version312b, MoPropertyMeta.READ_ONLY, None, None, None, None, [], []),
        "status": MoPropertyMeta("status", "status", "string", VersionMeta.Version312b, MoPropertyMeta.READ_WRITE, 0x80, None, None, r"""((removed|created|modified|deleted),){0,3}(removed|created|modified|deleted){0,1}""", [], []),
    }

    prop_map = {
        "blockSize": "block_size", 
        "childAction": "child_action", 
        "connectionProtocol": "connection_protocol", 
        "descr": "descr", 
        "deviceType": "device_type", 
        "dn": "dn", 
        "intId": "int_id", 
        "name": "name", 
        "numberOfBlocks": "number_of_blocks", 
        "policyLevel": "policy_level", 
        "policyOwner": "policy_owner", 
        "rn": "rn", 
        "sacl": "sacl", 
        "size": "size", 
        "status": "status", 
    }

    def __init__(self, parent_mo_or_dn, **kwargs):
        self._dirty_mask = 0
        self.block_size = None
        self.child_action = None
        self.connection_protocol = None
        self.descr = None
        self.device_type = None
        self.int_id = None
        self.name = None
        self.number_of_blocks = None
        self.policy_level = None
        self.policy_owner = None
        self.sacl = None
        self.size = None
        self.status = None

        ManagedObject.__init__(self, "EquipmentEmbeddedStorageDef", parent_mo_or_dn, **kwargs)



"""This module contains the general information for IqnpoolBlock ManagedObject."""

from ...ucsmo import ManagedObject
from ...ucscoremeta import MoPropertyMeta, MoMeta
from ...ucsmeta import VersionMeta


class IqnpoolBlockConsts:
    pass


class IqnpoolBlock(ManagedObject):
    """This is IqnpoolBlock class."""

    consts = IqnpoolBlockConsts()
    naming_props = set(['suffix', 'from', 'to'])

    mo_meta = MoMeta("IqnpoolBlock", "iqnpoolBlock", "block-[suffix]-from-[r_from]-to-[to]", VersionMeta.Version201m, "InputOutput", 0xff, [], ["admin", "ls-storage-policy"], ['iqnpoolPool'], [], ["Add", "Get", "Remove"])

    prop_meta = {
        "child_action": MoPropertyMeta("child_action", "childAction", "string", VersionMeta.Version201m, MoPropertyMeta.INTERNAL, 0x2, None, None, r"""((deleteAll|ignore|deleteNonPresent),){0,2}(deleteAll|ignore|deleteNonPresent){0,1}""", [], []),
        "dn": MoPropertyMeta("dn", "dn", "string", VersionMeta.Version201m, MoPropertyMeta.READ_ONLY, 0x4, 0, 256, None, [], []),
        "r_from": MoPropertyMeta("r_from", "from", "ushort", VersionMeta.Version201m, MoPropertyMeta.NAMING, 0x8, None, None, None, [], ["0-65535"]),
        "rn": MoPropertyMeta("rn", "rn", "string", VersionMeta.Version201m, MoPropertyMeta.READ_ONLY, 0x10, 0, 256, None, [], []),
        "sacl": MoPropertyMeta("sacl", "sacl", "string", VersionMeta.Version302c, MoPropertyMeta.READ_ONLY, None, None, None, r"""((none|del|mod|addchild|cascade),){0,4}(none|del|mod|addchild|cascade){0,1}""", [], []),
        "status": MoPropertyMeta("status", "status", "string", VersionMeta.Version201m, MoPropertyMeta.READ_WRITE, 0x20, None, None, r"""((removed|created|modified|deleted),){0,3}(removed|created|modified|deleted){0,1}""", [], []),
        "suffix": MoPropertyMeta("suffix", "suffix", "string", VersionMeta.Version201m, MoPropertyMeta.NAMING, 0x40, None, None, r"""[0-9a-zA-Z\.:-]{0,64}$""", [], []),
        "to": MoPropertyMeta("to", "to", "ushort", VersionMeta.Version201m, MoPropertyMeta.NAMING, 0x80, None, None, None, [], ["0-65535"]),
    }

    prop_map = {
        "childAction": "child_action", 
        "dn": "dn", 
        "from": "r_from", 
        "rn": "rn", 
        "sacl": "sacl", 
        "status": "status", 
        "suffix": "suffix", 
        "to": "to", 
    }

    def __init__(self, parent_mo_or_dn, suffix, r_from, to, **kwargs):
        self._dirty_mask = 0
        self.suffix = suffix
        self.r_from = r_from
        self.to = to
        self.child_action = None
        self.sacl = None
        self.status = None

        ManagedObject.__init__(self, "IqnpoolBlock", parent_mo_or_dn, **kwargs)



"""This module contains the general information for AdaptorNwMgmtCap ManagedObject."""

from ...ucsmo import ManagedObject
from ...ucscoremeta import MoPropertyMeta, MoMeta
from ...ucsmeta import VersionMeta


class AdaptorNwMgmtCapConsts:
    API_MENLO = "menlo"
    API_NONE = "none"
    API_PALO = "palo"
    MGMT_TRANSPORT_L2 = "L2"
    MGMT_TRANSPORT_L3 = "L3"
    MODE_FULL = "full"
    MODE_PARTIAL = "partial"
    OPER_POWER_REQUIREMENT_FULL = "full"
    OPER_POWER_REQUIREMENT_NONE = "none"
    OPER_POWER_REQUIREMENT_STANDBY = "standby"
    REBOOT_ACTION_ON_DESTRUCTIVE_ADAPTOR = "adaptor"
    REBOOT_ACTION_ON_DESTRUCTIVE_HOST = "host"
    REBOOT_ACTION_ON_DESTRUCTIVE_NONE = "none"


class AdaptorNwMgmtCap(ManagedObject):
    """This is AdaptorNwMgmtCap class."""

    consts = AdaptorNwMgmtCapConsts()
    naming_props = set([])

    mo_meta = MoMeta("AdaptorNwMgmtCap", "adaptorNwMgmtCap", "nw-mgmt", VersionMeta.Version101e, "InputOutput", 0x3ff, [], ["read-only"], ['adaptorFruCapProvider'], [], ["Get"])

    prop_meta = {
        "api": MoPropertyMeta("api", "api", "string", VersionMeta.Version101e, MoPropertyMeta.READ_WRITE, 0x2, None, None, None, ["menlo", "none", "palo"], []),
        "child_action": MoPropertyMeta("child_action", "childAction", "string", VersionMeta.Version101e, MoPropertyMeta.INTERNAL, 0x4, None, None, r"""((deleteAll|ignore|deleteNonPresent),){0,2}(deleteAll|ignore|deleteNonPresent){0,1}""", [], []),
        "dn": MoPropertyMeta("dn", "dn", "string", VersionMeta.Version101e, MoPropertyMeta.READ_ONLY, 0x8, 0, 256, None, [], []),
        "mgmt_transport": MoPropertyMeta("mgmt_transport", "mgmtTransport", "string", VersionMeta.Version101e, MoPropertyMeta.READ_WRITE, 0x10, None, None, None, ["L2", "L3"], []),
        "mode": MoPropertyMeta("mode", "mode", "string", VersionMeta.Version101e, MoPropertyMeta.READ_WRITE, 0x20, None, None, None, ["full", "partial"], []),
        "oper_power_requirement": MoPropertyMeta("oper_power_requirement", "operPowerRequirement", "string", VersionMeta.Version101e, MoPropertyMeta.READ_WRITE, 0x40, None, None, None, ["full", "none", "standby"], []),
        "reboot_action_on_destructive": MoPropertyMeta("reboot_action_on_destructive", "rebootActionOnDestructive", "string", VersionMeta.Version101e, MoPropertyMeta.READ_WRITE, 0x80, None, None, None, ["adaptor", "host", "none"], []),
        "rn": MoPropertyMeta("rn", "rn", "string", VersionMeta.Version101e, MoPropertyMeta.READ_ONLY, 0x100, 0, 256, None, [], []),
        "sacl": MoPropertyMeta("sacl", "sacl", "string", VersionMeta.Version302c, MoPropertyMeta.READ_ONLY, None, None, None, r"""((none|del|mod|addchild|cascade),){0,4}(none|del|mod|addchild|cascade){0,1}""", [], []),
        "status": MoPropertyMeta("status", "status", "string", VersionMeta.Version101e, MoPropertyMeta.READ_WRITE, 0x200, None, None, r"""((removed|created|modified|deleted),){0,3}(removed|created|modified|deleted){0,1}""", [], []),
    }

    prop_map = {
        "api": "api", 
        "childAction": "child_action", 
        "dn": "dn", 
        "mgmtTransport": "mgmt_transport", 
        "mode": "mode", 
        "operPowerRequirement": "oper_power_requirement", 
        "rebootActionOnDestructive": "reboot_action_on_destructive", 
        "rn": "rn", 
        "sacl": "sacl", 
        "status": "status", 
    }

    def __init__(self, parent_mo_or_dn, **kwargs):
        self._dirty_mask = 0
        self.api = None
        self.child_action = None
        self.mgmt_transport = None
        self.mode = None
        self.oper_power_requirement = None
        self.reboot_action_on_destructive = None
        self.sacl = None
        self.status = None

        ManagedObject.__init__(self, "AdaptorNwMgmtCap", parent_mo_or_dn, **kwargs)



"""This module contains the general information for BiosVfUSBSystemIdlePowerOptimizingSetting ManagedObject."""

from ...ucsmo import ManagedObject
from ...ucscoremeta import MoPropertyMeta, MoMeta
from ...ucsmeta import VersionMeta


class BiosVfUSBSystemIdlePowerOptimizingSettingConsts:
    SUPPORTED_BY_DEFAULT_NO = "no"
    SUPPORTED_BY_DEFAULT_YES = "yes"
    VP_USBIDLE_POWER_OPTIMIZING_HIGH_PERFORMANCE = "high-performance"
    VP_USBIDLE_POWER_OPTIMIZING_LOWER_IDLE_POWER = "lower-idle-power"
    VP_USBIDLE_POWER_OPTIMIZING_PLATFORM_DEFAULT = "platform-default"
    VP_USBIDLE_POWER_OPTIMIZING_PLATFORM_RECOMMENDED = "platform-recommended"


class BiosVfUSBSystemIdlePowerOptimizingSetting(ManagedObject):
    """This is BiosVfUSBSystemIdlePowerOptimizingSetting class."""

    consts = BiosVfUSBSystemIdlePowerOptimizingSettingConsts()
    naming_props = set([])

    mo_meta = MoMeta("BiosVfUSBSystemIdlePowerOptimizingSetting", "biosVfUSBSystemIdlePowerOptimizingSetting", "USB-System-Idle-Power-Optimizing-Setting", VersionMeta.Version142b, "InputOutput", 0x3f, [], ["admin", "ls-compute", "ls-config", "ls-server", "ls-server-policy", "pn-policy"], ['biosSettings', 'biosVProfile'], [], ["Get", "Set"])

    prop_meta = {
        "child_action": MoPropertyMeta("child_action", "childAction", "string", VersionMeta.Version142b, MoPropertyMeta.INTERNAL, 0x2, None, None, r"""((deleteAll|ignore|deleteNonPresent),){0,2}(deleteAll|ignore|deleteNonPresent){0,1}""", [], []),
        "dn": MoPropertyMeta("dn", "dn", "string", VersionMeta.Version142b, MoPropertyMeta.READ_ONLY, 0x4, 0, 256, None, [], []),
        "prop_acl": MoPropertyMeta("prop_acl", "propAcl", "ulong", VersionMeta.Version302c, MoPropertyMeta.READ_ONLY, None, None, None, None, [], []),
        "rn": MoPropertyMeta("rn", "rn", "string", VersionMeta.Version142b, MoPropertyMeta.READ_ONLY, 0x8, 0, 256, None, [], []),
        "sacl": MoPropertyMeta("sacl", "sacl", "string", VersionMeta.Version302c, MoPropertyMeta.READ_ONLY, None, None, None, r"""((none|del|mod|addchild|cascade),){0,4}(none|del|mod|addchild|cascade){0,1}""", [], []),
        "status": MoPropertyMeta("status", "status", "string", VersionMeta.Version142b, MoPropertyMeta.READ_WRITE, 0x10, None, None, r"""((removed|created|modified|deleted),){0,3}(removed|created|modified|deleted){0,1}""", [], []),
        "supported_by_default": MoPropertyMeta("supported_by_default", "supportedByDefault", "string", VersionMeta.Version302c, MoPropertyMeta.READ_ONLY, None, None, None, None, ["no", "yes"], []),
        "vp_usb_idle_power_optimizing": MoPropertyMeta("vp_usb_idle_power_optimizing", "vpUSBIdlePowerOptimizing", "string", VersionMeta.Version142b, MoPropertyMeta.READ_WRITE, 0x20, None, None, None, ["high-performance", "lower-idle-power", "platform-default", "platform-recommended"], []),
    }

    prop_map = {
        "childAction": "child_action", 
        "dn": "dn", 
        "propAcl": "prop_acl", 
        "rn": "rn", 
        "sacl": "sacl", 
        "status": "status", 
        "supportedByDefault": "supported_by_default", 
        "vpUSBIdlePowerOptimizing": "vp_usb_idle_power_optimizing", 
    }

    def __init__(self, parent_mo_or_dn, **kwargs):
        self._dirty_mask = 0
        self.child_action = None
        self.prop_acl = None
        self.sacl = None
        self.status = None
        self.supported_by_default = None
        self.vp_usb_idle_power_optimizing = None

        ManagedObject.__init__(self, "BiosVfUSBSystemIdlePowerOptimizingSetting", parent_mo_or_dn, **kwargs)



"""This module contains the general information for StorageIniGroup ManagedObject."""

from ...ucsmo import ManagedObject
from ...ucscoremeta import MoPropertyMeta, MoMeta
from ...ucsmeta import VersionMeta


class StorageIniGroupConsts:
    INT_ID_NONE = "none"
    OPER_PROTOCOL_DERIVED = "derived"
    OPER_PROTOCOL_FC = "fc"
    OPER_PROTOCOL_ISCSI = "iscsi"
    OPER_STATE_MISCONFIGURED = "misconfigured"
    OPER_STATE_OK = "ok"
    OWNER_CONN_POLICY = "conn_policy"
    OWNER_INITIATOR_POLICY = "initiator_policy"
    OWNER_LOGICAL = "logical"
    OWNER_PHYSICAL = "physical"
    OWNER_POLICY = "policy"
    OWNER_UNKNOWN = "unknown"
    POLICY_OWNER_LOCAL = "local"
    POLICY_OWNER_PENDING_POLICY = "pending-policy"
    POLICY_OWNER_POLICY = "policy"
    PROTOCOL_DERIVED = "derived"
    PROTOCOL_FC = "fc"
    PROTOCOL_ISCSI = "iscsi"


class StorageIniGroup(ManagedObject):
    """This is StorageIniGroup class."""

    consts = StorageIniGroupConsts()
    naming_props = set(['name'])

    mo_meta = MoMeta("StorageIniGroup", "storageIniGroup", "grp-[name]", VersionMeta.Version211a, "InputOutput", 0xfff, [], ["admin", "ls-compute", "ls-config", "ls-server", "ls-storage"], ['lsServer', 'vnicSanConnPolicy'], ['faultInst', 'storageInitiator', 'vnicFcGroupDef'], ["Add", "Get", "Remove", "Set"])

    prop_meta = {
        "child_action": MoPropertyMeta("child_action", "childAction", "string", VersionMeta.Version211a, MoPropertyMeta.INTERNAL, 0x2, None, None, r"""((deleteAll|ignore|deleteNonPresent),){0,2}(deleteAll|ignore|deleteNonPresent){0,1}""", [], []),
        "descr": MoPropertyMeta("descr", "descr", "string", VersionMeta.Version211a, MoPropertyMeta.READ_WRITE, 0x4, None, None, r"""[ !#$%&\(\)\*\+,\-\./:;\?@\[\]_\{\|\}~a-zA-Z0-9]{0,256}""", [], []),
        "dn": MoPropertyMeta("dn", "dn", "string", VersionMeta.Version211a, MoPropertyMeta.READ_ONLY, 0x8, 0, 256, None, [], []),
        "group_policy_name": MoPropertyMeta("group_policy_name", "groupPolicyName", "string", VersionMeta.Version211a, MoPropertyMeta.READ_WRITE, 0x10, None, None, r"""[\-\.:_a-zA-Z0-9]{0,16}""", [], []),
        "int_id": MoPropertyMeta("int_id", "intId", "string", VersionMeta.Version211a, MoPropertyMeta.INTERNAL, None, None, None, None, ["none"], ["0-4294967295"]),
        "name": MoPropertyMeta("name", "name", "string", VersionMeta.Version211a, MoPropertyMeta.NAMING, 0x20, None, None, r"""[\-\.:_a-zA-Z0-9]{1,16}""", [], []),
        "oper_protocol": MoPropertyMeta("oper_protocol", "operProtocol", "string", VersionMeta.Version211a, MoPropertyMeta.READ_ONLY, None, None, None, None, ["derived", "fc", "iscsi"], []),
        "oper_state": MoPropertyMeta("oper_state", "operState", "string", VersionMeta.Version212a, MoPropertyMeta.READ_ONLY, None, None, None, None, ["misconfigured", "ok"], []),
        "owner": MoPropertyMeta("owner", "owner", "string", VersionMeta.Version211a, MoPropertyMeta.READ_ONLY, None, None, None, None, ["conn_policy", "initiator_policy", "logical", "physical", "policy", "unknown"], []),
        "policy_level": MoPropertyMeta("policy_level", "policyLevel", "uint", VersionMeta.Version211a, MoPropertyMeta.READ_ONLY, None, None, None, None, [], []),
        "policy_name": MoPropertyMeta("policy_name", "policyName", "string", VersionMeta.Version211a, MoPropertyMeta.READ_WRITE, 0x40, None, None, r"""[\-\.:_a-zA-Z0-9]{0,16}""", [], []),
        "policy_owner": MoPropertyMeta("policy_owner", "policyOwner", "string", VersionMeta.Version211a, MoPropertyMeta.READ_WRITE, 0x80, None, None, None, ["local", "pending-policy", "policy"], []),
        "protocol": MoPropertyMeta("protocol", "protocol", "string", VersionMeta.Version211a, MoPropertyMeta.CREATE_ONLY, 0x100, None, None, None, ["derived", "fc", "iscsi"], []),
        "rmt_disk_cfg_name": MoPropertyMeta("rmt_disk_cfg_name", "rmtDiskCfgName", "string", VersionMeta.Version211a, MoPropertyMeta.READ_WRITE, 0x200, None, None, r"""[\-\.:_a-zA-Z0-9]{0,16}""", [], []),
        "rn": MoPropertyMeta("rn", "rn", "string", VersionMeta.Version211a, MoPropertyMeta.READ_ONLY, 0x400, 0, 256, None, [], []),
        "sacl": MoPropertyMeta("sacl", "sacl", "string", VersionMeta.Version302c, MoPropertyMeta.READ_ONLY, None, None, None, r"""((none|del|mod|addchild|cascade),){0,4}(none|del|mod|addchild|cascade){0,1}""", [], []),
        "status": MoPropertyMeta("status", "status", "string", VersionMeta.Version211a, MoPropertyMeta.READ_WRITE, 0x800, None, None, r"""((removed|created|modified|deleted),){0,3}(removed|created|modified|deleted){0,1}""", [], []),
    }

    prop_map = {
        "childAction": "child_action", 
        "descr": "descr", 
        "dn": "dn", 
        "groupPolicyName": "group_policy_name", 
        "intId": "int_id", 
        "name": "name", 
        "operProtocol": "oper_protocol", 
        "operState": "oper_state", 
        "owner": "owner", 
        "policyLevel": "policy_level", 
        "policyName": "policy_name", 
        "policyOwner": "policy_owner", 
        "protocol": "protocol", 
        "rmtDiskCfgName": "rmt_disk_cfg_name", 
        "rn": "rn", 
        "sacl": "sacl", 
        "status": "status", 
    }

    def __init__(self, parent_mo_or_dn, name, **kwargs):
        self._dirty_mask = 0
        self.name = name
        self.child_action = None
        self.descr = None
        self.group_policy_name = None
        self.int_id = None
        self.oper_protocol = None
        self.oper_state = None
        self.owner = None
        self.policy_level = None
        self.policy_name = None
        self.policy_owner = None
        self.protocol = None
        self.rmt_disk_cfg_name = None
        self.sacl = None
        self.status = None

        ManagedObject.__init__(self, "StorageIniGroup", parent_mo_or_dn, **kwargs)



"""This module contains the meta information of AaaGetNComputeAuthTokenInternalByDn ExternalMethod."""

from ..ucscoremeta import MethodMeta, MethodPropertyMeta

method_meta = MethodMeta("AaaGetNComputeAuthTokenInternalByDn", "aaaGetNComputeAuthTokenInternalByDn", "Version142b")

prop_meta = {
    "cookie": MethodPropertyMeta("Cookie", "cookie", "Xs:string", "Version142b", "InputOutput", False),
    "in_dn": MethodPropertyMeta("InDn", "inDn", "Xs:string", "Version142b", "Input", False),
    "in_id": MethodPropertyMeta("InId", "inId", "Xs:unsignedInt", "Version142b", "Input", False),
    "in_ipv4": MethodPropertyMeta("InIpv4", "inIpv4", "Xs:string", "Version142b", "Input", False),
    "in_locales": MethodPropertyMeta("InLocales", "inLocales", "Xs:string", "Version142b", "Input", False),
    "in_number_of": MethodPropertyMeta("InNumberOf", "inNumberOf", "Xs:unsignedByte", "Version142b", "Input", False),
    "in_parent_sess": MethodPropertyMeta("InParentSess", "inParentSess", "Xs:string", "Version142b", "Input", False),
    "in_priv": MethodPropertyMeta("InPriv", "inPriv", "Xs:string", "Version142b", "Input", False),
    "in_remote": MethodPropertyMeta("InRemote", "inRemote", "Xs:string", "Version142b", "Input", False),
    "in_role_list": MethodPropertyMeta("InRoleList", "inRoleList", "Xs:string", "Version142b", "Input", False),
    "in_user": MethodPropertyMeta("InUser", "inUser", "Xs:string", "Version142b", "Input", False),
    "out_out_user": MethodPropertyMeta("OutOutUser", "outOutUser", "Xs:string", "Version142b", "Output", False),
    "out_tokens": MethodPropertyMeta("OutTokens", "outTokens", "Xs:string", "Version142b", "Output", False),
}

prop_map = {
    "cookie": "cookie",
    "inDn": "in_dn",
    "inId": "in_id",
    "inIpv4": "in_ipv4",
    "inLocales": "in_locales",
    "inNumberOf": "in_number_of",
    "inParentSess": "in_parent_sess",
    "inPriv": "in_priv",
    "inRemote": "in_remote",
    "inRoleList": "in_role_list",
    "inUser": "in_user",
    "outOutUser": "out_out_user",
    "outTokens": "out_tokens",
}
